import unittest
import typing
from pydantic import BaseModel, ValidationError

from hololinked.utils import (
    get_input_model_from_signature,
    issubklass,
    pydantic_validate_args_kwargs,
    json_schema_merge_args_to_kwargs,
)

try:
    from .utils import TestCase, TestRunner
except ImportError:
    from utils import TestCase, TestRunner


class TestUtils(TestCase):
    def test_1_pydantic_function_signature_validation(self):
        def func_without_args():
            return 1

        model = get_input_model_from_signature(func_without_args)
        self.assertTrue(model is None)

        """
        Test Sequence:
        1. Create model from function signature
        2. Check model annotations
        3. Check model fields length 
        4. Check model config (pydantic's model_config)
        5. Validation with correction and wrong invokation of function 
        6. Always check exception strings for ValueError
        7. Use ValidationError if pydantic is supposed to raise the Error
        """

        """
        Signatures that we will validate:
        1. func_with_annotations(a: int, b: int) -> int:
        2. func_with_missing_annotations(a: int, b):
        3. func_with_no_annotations(a, b):
        4. func_with_kwargs(a: int, b: int, **kwargs):
        5. func_with_annotated_kwargs(a: int, b: int, **kwargs: typing.Dict[str, int]):
        6. func_with_args(*args):
        7. func_with_annotated_args(*args: typing.List[int]):
        8. func_with_args_and_kwargs(*args, **kwargs):
        9. func_with_annotated_args_and_kwargs(*args: typing.List[int], **kwargs: typing.Dict[str, int]):
        10. func_with_positional_only_args(a, b, /):
        11. func_with_keyword_only_args(*, a, b):
        12. func_with_positional_only_args_and_kwargs(a, *args, b, **kwargs):
        """

        ####################
        ##### create model from function signature
        # 1. func_with_annotations(a: int, b: int) -> int:
        def func_with_annotations(a: int, b: int) -> int:
            return a + b

        model = get_input_model_from_signature(func_with_annotations)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["a"].annotation, int)
        self.assertEqual(model.model_fields["b"].annotation, int)
        self.assertEqual(len(model.model_fields), 2)
        self.assertEqual(model.model_config["extra"], "forbid")
        ##### validate correct usage
        # For all the following cases, see block comment below the test case for details
        # 1. correct usage with keyword arguments
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2})
        # 2. incorrect argument types with keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": "2"})
        # 3. missing keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1})
        # 4. too many keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        # 5. correct usage with positional arguments
        pydantic_validate_args_kwargs(model, args=(1, 2))
        # 6. incorrect argument types with positional arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(1, "2"))
        # 7. too many positional arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2, 3))
        self.assertTrue(str(ex.exception).startswith("Too many positional arguments"))
        # 8. missing positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,))
        # 9. correct usage with positional and keyword arguments
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2})
        # 10. incorrect ordering with positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,), kwargs={"a": 2})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        # 11. incorrect usage with both positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=("1", 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        ####################

        # 1. correct usage with keyword arguments
        # 2. incorrect argument types with keyword arguments
        # 3. missing keyword arguments
        # 4. too many keyword arguments
        # 5. correct usage with positional arguments
        # 6. incorrect argument types with positional arguments
        # 7. too many positional arguments
        # 8. missing positional arguments
        # 9. correct usage with positional and keyword arguments
        # 10. incorrect ordering with positional and keyword arguments
        # 11. additional cases of incorrect usage falling under the same categories

        ####################
        ##### create model from function signature
        # 2. func_with_missing_annotations(a: int, b):
        def func_with_missing_annotations(a: int, b):
            return a + b

        model = get_input_model_from_signature(func_with_missing_annotations)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["a"].annotation, int)
        self.assertEqual(model.model_fields["b"].annotation, typing.Any)
        self.assertEqual(len(model.model_fields), 2)
        self.assertEqual(model.model_config["extra"], "forbid")
        ##### validate correct usage
        # 1. correct usage with keyword arguments
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2})
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": "2"})
        pydantic_validate_args_kwargs(model, kwargs={"a": 2, "b": list()})
        # 2. incorrect argument types with keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": "1", "b": "2"})
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": list(), "b": dict()})
        # 3. missing keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1})
        # 4. too many keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        # 5. correct positional arguments
        pydantic_validate_args_kwargs(model, args=(1, 2))
        pydantic_validate_args_kwargs(model, args=(1, "2"))
        pydantic_validate_args_kwargs(model, args=(2, list()))
        # 6. incorrect argument types with positional arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=("1", "2"))
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(list(), dict()))
        # 7. too many positional arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2, 3))
        self.assertTrue(str(ex.exception).startswith("Too many positional arguments"))
        # 8. missing positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,))
        # 9. correct usage with positional and keyword arguments
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2})
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": "2"})
        pydantic_validate_args_kwargs(model, args=(2,), kwargs={"b": list()})
        # 10. incorrect ordering with positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,), kwargs={"a": 2})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        # 11. incorrect usage with both positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=("1", 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        ####################

        ####################
        ##### create model from function signature
        # 3. func_with_no_annotations(a, b):
        def func_with_no_annotations(a, b):
            return a + b

        model = get_input_model_from_signature(func_with_no_annotations, model_for_empty_annotations=True)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["a"].annotation, typing.Any)
        self.assertEqual(model.model_fields["b"].annotation, typing.Any)
        self.assertEqual(len(model.model_fields), 2)
        self.assertEqual(model.model_config["extra"], "forbid")
        ##### validate correct usage
        # 1. correct usage
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2})
        pydantic_validate_args_kwargs(model, kwargs={"a": 1.2, "b": "2"})
        pydantic_validate_args_kwargs(model, kwargs={"a": dict(), "b": list()})
        # 2. incorrect argument types
        # typing.Any allows any type, so no ValidationError
        # 3. missing keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": list()})
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"b": dict()})
        # 4. too many keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        # 5. correct positional arguments
        pydantic_validate_args_kwargs(model, args=(1, 2))
        pydantic_validate_args_kwargs(model, args=(1, "2"))
        pydantic_validate_args_kwargs(model, args=(dict(), list()))
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2})
        # 6. incorrect argument types with positional arguments
        # typing.Any allows any type, so no ValidationError
        # 7. too many positional arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2, 3))
        self.assertTrue(str(ex.exception).startswith("Too many positional arguments"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(dict(), list(), 3))
        self.assertTrue(str(ex.exception).startswith("Too many positional arguments"))
        # 8. missing positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,))
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(dict(),))
        # 9. correct usage with positional and keyword arguments
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2})
        pydantic_validate_args_kwargs(model, args=(1.1,), kwargs={"b": "2"})
        pydantic_validate_args_kwargs(model, args=(dict(),), kwargs={"b": list()})
        # 10. incorrect ordering with positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,), kwargs={"a": 2})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        # 11. incorrect usage with both positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=("1", 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))

        model = get_input_model_from_signature(func_with_no_annotations)
        self.assertTrue(model is None)

        ####################
        ##### create model from function signature
        # 4. func_with_kwargs(a: int, b: int, **kwargs):
        def func_with_kwargs(a: int, b: int, **kwargs):
            return a + b

        model = get_input_model_from_signature(func_with_kwargs)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["a"].annotation, int)
        self.assertEqual(model.model_fields["b"].annotation, int)
        self.assertEqual(len(model.model_fields), 3)
        self.assertEqual(model.model_config["extra"], "forbid")
        ##### validate correct usage
        # 1. correct usage
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2})
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": 3})
        pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"c": "3"})
        # 2. incorrect argument types
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": "2"})
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": "2", "c": "3"})
        # 3. missing keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1})
        # 4. too many keyword arguments
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": 3, "d": 4})  # OK, not an error
        # 5. correct positional arguments
        pydantic_validate_args_kwargs(model, args=(1, 2))
        # 6. incorrect argument types with positional arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(1, "2"))
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=("1", 2))
        # 7. too many positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2, 3))
        # 8. missing positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,))
        # 9. correct usage with positional and keyword arguments
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2})
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2, "c": 3})
        # 10. incorrect ordering with positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,), kwargs={"a": 2})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"a": 3})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"b": 3})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        # 11. incorrect usage with both positional and keyword arguments
        # any extra keyword argument is allowed

        ####################
        ##### create model from function signature
        # 5. func_with_annotated_kwargs(a: int, b: int, **kwargs: typing.Dict[str, int]):
        def func_with_annotated_kwargs(a: int, b: int, **kwargs: typing.Dict[str, int]):
            return a + b

        model = get_input_model_from_signature(func_with_annotated_kwargs)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["a"].annotation, int)
        self.assertEqual(model.model_fields["b"].annotation, int)
        self.assertEqual(model.model_fields["kwargs"].annotation, typing.Dict[str, int])
        self.assertEqual(len(model.model_fields), 3)
        self.assertEqual(model.model_config["extra"], "forbid")
        # 1. correct usage
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2})
        pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": 3})
        pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"c": 3})
        # 2. incorrect argument types
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": "2"})
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": "3"})
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1, "b": 2, "c": list()})
        # 3. missing keyword arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1})
        # 4. too many keyword arguments
        # OK, not an error
        # 5. correct positional arguments
        pydantic_validate_args_kwargs(model, args=(1, 2))
        # 6. incorrect argument types with positional arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(1, "2"))
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(dict(), 2))
        # 7. too many positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2, 3))
        # 8. missing positional arguments
        with self.assertRaises(ValidationError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,))
        # 9. correct usage with positional and keyword arguments
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2})
        pydantic_validate_args_kwargs(model, args=(1,), kwargs={"b": 2, "c": 3})
        # 10. incorrect ordering with positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,), kwargs={"a": 2})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"a": 3})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"b": 3})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"a": list(), "c": 3})
        self.assertTrue(str(ex.exception).startswith("Multiple values for argument"))
        # 11. incorrect usage with both positional and keyword arguments
        # any extra keyword argument is allowed so long it is of type int

        # both the following are not allowed in python - its also illogical
        # def func_with_double_args(*args1, *args2):
        #     """syntax error"""
        #     return
        # def func_with_double_kwargs(**kwargs1, **kwargs2):
        #     """syntax error"""
        #     return

        ####################
        ##### create model from function signature
        # 6. func_with_args(*args):
        def func_with_args(*args):
            return sum(args)

        model = get_input_model_from_signature(func_with_args, model_for_empty_annotations=True)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["args"].annotation, typing.Tuple)
        self.assertEqual(len(model.model_fields), 1)
        self.assertEqual(model.model_config["extra"], "forbid")
        # 1. correct usage
        pydantic_validate_args_kwargs(model, args=(1, 2))
        pydantic_validate_args_kwargs(model)
        pydantic_validate_args_kwargs(model, args=(dict()))
        # 2. incorrect argument types
        # OK, since args is a tuple of any type
        # 3. missing keyword arguments
        # OK, since args is a tuple
        # 4. too many keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, kwargs={"a": 1})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        # 5. correct positional arguments
        pydantic_validate_args_kwargs(model, args=(1, 2))
        # 6. incorrect argument types with positional arguments
        # OK, since args is a tuple of any type
        # 7. too many positional arguments
        # OK, since args is a tuple of any length
        # 8. missing positional arguments
        # OK, since args is a tuple of any length
        # 9. correct usage with positional and keyword arguments
        # no keyword arguments
        # 10. incorrect ordering with positional and keyword arguments
        # OK, since args is a tuple and not keywords, no multiple values
        # 11. incorrect usage with both positional and keyword arguments
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1,), kwargs={"a": 2})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        with self.assertRaises(ValueError) as ex:
            pydantic_validate_args_kwargs(model, args=(1, 2), kwargs={"c": 3})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))

        model = get_input_model_from_signature(func_with_args)
        self.assertTrue(model is None)
        ####################

        ####################
        ##### create model from function signature
        # 7. func_with_annotated_args(*args: typing.List[int]):
        def func_with_annotated_args(*args: typing.List[int]):
            return sum(args)

        model = get_input_model_from_signature(func_with_annotated_args)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["args"].annotation, typing.List[int])
        self.assertEqual(len(model.model_fields), 1)
        self.assertEqual(model.model_config["extra"], "forbid")
        # 1. correct usage with keyword arguments
        # not possible, since args is a tuple
        # 2. incorrect argument types with keyword arguments
        # keyword arguments are not allowed
        # 3. missing keyword arguments
        # not possible
        # 4. too many keyword arguments
        with self.assertRaises(ValueError):
            pydantic_validate_args_kwargs(model, kwargs={"a": 1})
        self.assertTrue(str(ex.exception).startswith("Unexpected keyword arguments"))
        # 5. correct usage with positional arguments
        pydantic_validate_args_kwargs(model)
        pydantic_validate_args_kwargs(model, args=(1, 2))
        # 6. incorrect argument types with positional arguments
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(1, "2"))
        with self.assertRaises(ValidationError):
            pydantic_validate_args_kwargs(model, args=(dict(),))
        # 7. too many positional arguments
        # OK, since args is a list of any length
        # 8. missing positional arguments
        # OK, since args is a list of any length
        # 9. correct usage with positional and keyword arguments
        # not possible
        # 10. incorrect ordering with positional and keyword arguments
        # OK, since args is a list and not keywords, no multiple values
        # 11. incorrect usage with both positional and keyword arguments
        # not possible

        #####################
        ##### create model from function signature
        # 8. func_with_args_and_kwargs(*args, **kwargs):
        def func_with_args_and_kwargs(*args, **kwargs):
            return sum(args) + sum(kwargs.values())

        # no model
        model = get_input_model_from_signature(func_with_args_and_kwargs)
        self.assertTrue(model is None)
        # check model for empty annotations
        model = get_input_model_from_signature(func_with_args_and_kwargs, model_for_empty_annotations=True)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["args"].annotation, typing.Tuple)
        self.assertEqual(model.model_fields["kwargs"].annotation, typing.Dict[str, typing.Any])
        self.assertEqual(len(model.model_fields), 2)
        self.assertEqual(model.model_config["extra"], "forbid")

        def func_with_annotated_args_and_kwargs(*args: typing.List[int], **kwargs: typing.Dict[str, int]):
            return sum(args) + sum(kwargs.values())

        model = get_input_model_from_signature(func_with_annotated_args_and_kwargs)
        self.assertTrue(issubklass(model, BaseModel))
        self.assertEqual(model.model_fields["args"].annotation, typing.List[int])
        self.assertEqual(model.model_fields["kwargs"].annotation, typing.Dict[str, int])
        self.assertEqual(len(model.model_fields), 2)
        self.assertEqual(model.model_config["extra"], "forbid")


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())



import jsonschema
from pydantic import BaseModel

from ..utils import pydantic_validate_args_kwargs, json_schema_merge_args_to_kwargs
from ..constants import JSON


class BaseSchemaValidator:  # type definition
    """
    Base class for all schema validators.
    Serves as a type definition.
    """

    def __init__(self, schema: JSON | BaseModel) -> None:
        self.schema = schema

    def validate(self, data) -> None:
        """
        validate the data against the schema.
        """
        raise NotImplementedError("validate method must be implemented by subclass")

    def validate_method_call(self, args, kwargs) -> None:
        """
        validate the method call against the schema.
        """
        raise NotImplementedError("validate_method_call method must be implemented by subclass")


class JSONSchemaValidator(BaseSchemaValidator):
    """
    JSON schema validator according to standard python JSON schema.
    Somewhat slow, consider msgspec if possible.
    """

    def __init__(self, schema) -> None:
        jsonschema.Draft7Validator.check_schema(schema)
        super().__init__(schema)
        self.validator = jsonschema.Draft7Validator(schema)

    def validate(self, data) -> None:
        self.validator.validate(data)

    def validate_method_call(self, args, kwargs) -> None:
        if len(args) > 0:
            kwargs = json_schema_merge_args_to_kwargs(self.schema, args, kwargs)
        self.validate(kwargs)

    def json(self) -> JSON:
        """allows JSON (de-)serializable of the instance itself"""
        return self.schema

    def __get_state__(self):
        return self.schema

    def __set_state__(self, schema):
        return JSONSchemaValidator(schema)


class PydanticSchemaValidator(BaseSchemaValidator):
    """
    JSON schema validator according to pydantic.
    """

    def __init__(self, schema: BaseModel) -> None:
        super().__init__(schema)
        self.validator = schema.model_validate

    def validate(self, data) -> None:
        self.validator(data)

    def validate_method_call(self, args, kwargs) -> None:
        pydantic_validate_args_kwargs(self.schema, args, kwargs)

    def json(self) -> JSON:
        """allows JSON (de-)serializable of the instance itself"""
        return self.schema.model_dump_json()

    def __get_state__(self):
        return self.json()

    def __set_state__(self, schema: JSON):
        return PydanticSchemaValidator(BaseModel(**schema))


try:
    import fastjsonschema

    class FastJsonSchemaValidator(BaseSchemaValidator):
        """
        JSON schema validator according to fast JSON schema.
        Useful for performance with dictionary based schema specification
        which msgspec has no built in support. Normally, for speed,
        one should try to use msgspec's struct concept.
        """

        def __init__(self, schema: JSON) -> None:
            super().__init__(schema)
            self.validator = fastjsonschema.compile(schema)

        def validate(self, data) -> None:
            """validates and raises exception when failed directly to the caller"""
            self.validator(data)

        def validate_method_call(self, args, kwargs) -> None:
            if len(args) > 0:
                kwargs = json_schema_merge_args_to_kwargs(self.schema, args, kwargs)
            self.validate(kwargs)

        def json(self) -> JSON:
            """allows JSON (de-)serializable of the instance itself"""
            return self.schema

        def __get_state__(self):
            return self.schema

        def __set_state__(self, schema):
            return FastJsonSchemaValidator(schema)

except ImportError:
    pass



# an end to end through the zmq object proxy client with IPC protocol which is assumed to be most stable
import time
import unittest
import logging
from uuid import uuid4
from hololinked.client.abstractions import SSE
from hololinked.client.factory import ClientFactory
from hololinked.client.proxy import ObjectProxy

try:
    from .things import TestThing
    from .utils import TestCase, TestRunner, fake, AsyncTestCase
except ImportError:
    from things import TestThing
    from utils import TestCase, TestRunner, fake, AsyncTestCase


class TestRPCEndToEnd(TestCase):
    """Test the zmq object proxy client"""

    @classmethod
    def setUpClass(cls):
        cls.thing_id = f"test-thing-{uuid4().hex[:8]}"
        cls.server_id = cls.thing_id
        super().setUpClass()
        cls.setUpThing()
        print("Test ZMQ IPC End to End")

    @classmethod
    def setUpThing(cls):
        """Set up the thing for the zmq object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_zmq_server(forked=True)
        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def tearDownClass(cls):
        """Test the stop of the zmq object proxy client"""
        cls._client = None
        cls.thing.rpc_server.stop()
        super().tearDownClass()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.zmq(
                cls.server_id,
                cls.thing_id,
                "IPC",
                log_level=logging.ERROR + 10,
                ignore_TD_errors=True,
            )
            return cls._client

    def test_01_creation_and_handshake(self):
        """Test the creation and handshake of the zmq object proxy client"""
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        self.assertTrue(
            len(thing.properties) + len(thing.actions) + len(thing.events)
            >= len(self.thing_model["properties"]) + len(self.thing_model["actions"]) + len(self.thing_model["events"])
        )

    def test_02_invoke_action(self):
        """Test the invocation of an action on the zmq object proxy client"""
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        # Test invoke_action method with reply
        self.assertEqual(thing.invoke_action("action_echo", fake.text(max_nb_chars=100)), fake.last)
        self.assertEqual(thing.invoke_action("action_echo", fake.sentence()), fake.last)
        self.assertEqual(thing.invoke_action("action_echo", fake.json()), fake.last)
        # Test invoke_action with dot notation
        self.assertEqual(thing.action_echo(fake.chrome()), fake.last)
        self.assertEqual(thing.action_echo(fake.sha256()), fake.last)
        self.assertEqual(thing.action_echo(fake.address()), fake.last)
        # Test invoke_action with no reply
        self.assertEqual(
            thing.invoke_action("set_non_remote_number_prop", fake.random_number(), oneway=True),
            None,
        )
        self.assertEqual(thing.get_non_remote_number_prop(), fake.last)
        # Test invoke_action in non blocking mode
        noblock_payload = fake.pylist(20, value_types=[int, float, str, bool])
        noblock_msg_id = thing.invoke_action("action_echo", noblock_payload, noblock=True)
        self.assertIsInstance(noblock_msg_id, str)
        self.assertEqual(
            thing.invoke_action("action_echo", fake.pylist(20, value_types=[int, float, str, bool])),
            fake.last,
        )
        self.assertEqual(
            thing.invoke_action("action_echo", fake.pylist(10, value_types=[int, float, str, bool])),
            fake.last,
        )
        self.assertEqual(thing.read_reply(noblock_msg_id), noblock_payload)

    def test_03_rwd_properties(self):
        """Test the read, write and delete of properties on the zmq object proxy client"""
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        # Test read_property method
        self.assertIsInstance(thing.read_property("number_prop"), (int, float))
        self.assertIsInstance(thing.read_property("string_prop"), str)
        self.assertIn(thing.read_property("selector_prop"), TestThing.selector_prop.objects)
        # Test write_property method
        thing.write_property("number_prop", fake.random_number())
        self.assertEqual(thing.read_property("number_prop"), fake.last)
        thing.write_property(
            "selector_prop",
            TestThing.selector_prop.objects[fake.random_int(0, len(TestThing.selector_prop.objects) - 1)],
        )
        self.assertEqual(
            thing.read_property("selector_prop"),
            TestThing.selector_prop.objects[fake.last],
        )
        thing.write_property("observable_list_prop", fake.pylist(25, value_types=[int, float, str, bool]))
        self.assertEqual(thing.read_property("observable_list_prop"), fake.last)
        # Test read property through dot notation attribute access
        self.assertIsInstance(thing.number_prop, (int, float))
        self.assertIsInstance(thing.string_prop, str)
        self.assertIn(thing.selector_prop, TestThing.selector_prop.objects)
        # Test write property through dot notation attribute access
        thing.number_prop = fake.random_number()
        self.assertEqual(thing.number_prop, fake.last)
        thing.selector_prop = TestThing.selector_prop.objects[
            fake.random_int(0, len(TestThing.selector_prop.objects) - 1)
        ]
        self.assertEqual(thing.selector_prop, TestThing.selector_prop.objects[fake.last])
        thing.observable_list_prop = fake.pylist(25, value_types=[int, float, str, bool])
        self.assertEqual(thing.observable_list_prop, fake.last)
        # Test one way write property
        thing.write_property("number_prop", fake.random_number(), oneway=True)
        self.assertEqual(thing.read_property("number_prop"), fake.last)
        thing.write_property(
            "selector_prop",
            TestThing.selector_prop.objects[fake.random_int(0, len(TestThing.selector_prop.objects) - 1)],
            oneway=True,
        )
        self.assertEqual(
            thing.read_property("selector_prop"),
            TestThing.selector_prop.objects[fake.last],
        )
        thing.write_property(
            "observable_list_prop",
            fake.pylist(25, value_types=[int, float, str, bool]),
            oneway=True,
        )
        self.assertEqual(thing.read_property("observable_list_prop"), fake.last)
        # Test noblock read property
        noblock_msg_id = thing.read_property("number_prop", noblock=True)
        self.assertIsInstance(noblock_msg_id, str)
        self.assertIn(thing.read_property("selector_prop"), TestThing.selector_prop.objects)
        self.assertIsInstance(thing.read_property("string_prop"), str)
        self.assertEqual(thing.read_reply(noblock_msg_id), thing.number_prop)
        # Test noblock write property
        noblock_msg_id = thing.write_property("number_prop", fake.random_number(), noblock=True)
        self.assertIsInstance(noblock_msg_id, str)
        self.assertEqual(thing.read_property("number_prop"), fake.last)  # noblock worked
        self.assertEqual(thing.read_reply(noblock_msg_id), None)
        # Test exception propagation to client
        thing.string_prop = "world"
        self.assertEqual(thing.string_prop, "world")
        with self.assertRaises(ValueError):
            thing.string_prop = "WORLD"
        with self.assertRaises(TypeError):
            thing.int_prop = "5"
        # Test non remote prop (non-)availability on client
        with self.assertRaises(AttributeError):
            thing.non_remote_number_prop

    def test_04_RW_multiple_properties(self):
        # TD is not well defined for this yet, although both client and server separately work.
        # Test partial list of read write properties
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        # Test read_multiple_properties method
        thing.write_multiple_properties(number_prop=15, string_prop="foobar")
        self.assertEqual(thing.number_prop, 15)
        self.assertEqual(thing.string_prop, "foobar")
        # check prop that was not set in multiple properties

        thing.int_prop = 5
        thing.selector_prop = "b"
        thing.number_prop = -15  # simply override
        props = thing.read_multiple_properties(names=["selector_prop", "int_prop", "number_prop", "string_prop"])
        self.assertEqual(props["selector_prop"], "b")
        self.assertEqual(props["int_prop"], 5)
        self.assertEqual(props["number_prop"], -15)
        self.assertEqual(props["string_prop"], "foobar")

    def test_05_subscribe_event(self):
        """Test the subscription to an event on the zmq object proxy client"""
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)

        results = []

        def cb(value: SSE):
            results.append(value)

        thing.subscribe_event("test_event", cb)
        time.sleep(1)  # wait for the subscription to be established

        thing.push_events()
        time.sleep(3)  # wait for the event to be pushed
        self.assertGreater(len(results), 0, "No events received")
        self.assertEqual(len(results), 100)
        thing.unsubscribe_event("test_event")

    def test_06_observe_properties(self):
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)

        # First check if an attribute is set on the object proxy
        self.assertIsNotNone(thing, "observable_list_prop_change_event")
        self.assertIsNotNone(thing, "observable_readonly_prop_change_event")

        # req 1 - observable events come due to writing a property
        propective_values = [
            [1, 2, 3, 4, 5],
            ["a", "b", "c", "d", "e"],
            [1, "a", 2, "b", 3],
        ]
        result = []
        attempt = 0

        def cb(value: SSE):
            nonlocal attempt, result
            self.assertEqual(value.data, propective_values[attempt])
            result.append(value)
            attempt += 1

        thing.observe_property("observable_list_prop", cb)
        time.sleep(3)
        # Calm down for event publisher to connect fully as there is no handshake for events
        for value in propective_values:
            thing.observable_list_prop = value

        for i in range(20):
            if attempt == len(propective_values):
                break
            # wait for the callback to be called
            time.sleep(0.1)
        thing.unobserve_property("observable_list_prop")

        for res in result:
            self.assertIn(res.data, propective_values)

        # # req 2 - observable events come due to reading a property
        propective_values = [1, 2, 3, 4, 5]
        result = []
        attempt = 0

        def cb(value: SSE):
            nonlocal attempt, result
            self.assertEqual(value.data, propective_values[attempt])
            result.append(value)
            attempt += 1

        thing.observe_property("observable_readonly_prop", cb)
        time.sleep(3)
        # Calm down for event publisher to connect fully as there is no handshake for events
        for _ in propective_values:
            thing.observable_readonly_prop  # read property through do notation access

        for i in range(20):
            if attempt == len(propective_values):
                break
            # wait for the callback to be called
            time.sleep(0.1)

        thing.unobserve_property("observable_readonly_prop")
        for res in result:
            self.assertIn(res.data, propective_values)


class TestRPCEndToEndAsync(AsyncTestCase):
    @classmethod
    def setUpClass(cls):
        cls.thing_id = f"test-thing-{uuid4().hex[:8]}"
        cls.server_id = cls.thing_id
        super().setUpClass()
        cls.setUpThing()

    @classmethod
    def setUpThing(cls):
        """Set up the thing for the zmq object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_zmq_server(forked=True)
        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def tearDownClass(cls):
        cls._client = None
        cls.thing.rpc_server.stop()
        super().tearDownClass()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.zmq(
                cls.server_id,
                cls.thing_id,
                "IPC",
                log_level=logging.ERROR + 10,
                ignore_TD_errors=True,
            )
            return cls._client

    async def test_01_creation_and_handshake(self):
        """Test the creation and handshake of the zmq object proxy client"""
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        self.assertTrue(
            len(thing.properties) + len(thing.actions) + len(thing.events)
            >= len(self.thing_model["properties"]) + len(self.thing_model["actions"]) + len(self.thing_model["events"])
        )

    async def test_02_invoke_action(self):
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        self.assertEqual(
            await thing.async_invoke_action("action_echo", fake.text(max_nb_chars=100)),
            fake.last,
        )
        self.assertEqual(await thing.async_invoke_action("action_echo", fake.sentence()), fake.last)
        self.assertEqual(await thing.async_invoke_action("action_echo", fake.json()), fake.last)

    async def test_03_rwd_properties(self):
        """Test the read, write and delete of properties on the zmq object proxy client"""
        thing = self.get_client()
        self.assertIsInstance(thing, ObjectProxy)
        # Test read_property method
        self.assertIsInstance(await thing.async_read_property("number_prop"), (int, float))
        self.assertIsInstance(await thing.async_read_property("string_prop"), str)
        self.assertIn(
            await thing.async_read_property("selector_prop"),
            TestThing.selector_prop.objects,
        )
        # Test write_property method
        await thing.async_write_property("number_prop", fake.random_number())
        self.assertEqual(await thing.async_read_property("number_prop"), fake.last)
        await thing.async_write_property(
            "selector_prop",
            TestThing.selector_prop.objects[fake.random_int(0, len(TestThing.selector_prop.objects) - 1)],
        )
        self.assertEqual(
            await thing.async_read_property("selector_prop"),
            TestThing.selector_prop.objects[fake.last],
        )
        await thing.async_write_property("observable_list_prop", fake.pylist(25, value_types=[int, float, str, bool]))
        self.assertEqual(await thing.async_read_property("observable_list_prop"), fake.last)
        # await complete_pending_tasks_in_current_loop_async()


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestRPCEndToEnd))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestRPCEndToEndAsync))
    return suite


if __name__ == "__main__":
    runner = TestRunner()
    runner.run(load_tests(unittest.TestLoader(), None, None))



import logging
import unittest

from hololinked.client import ClientFactory

try:
    from .test_11_rpc_e2e import TestRPCEndToEnd, TestRPCEndToEndAsync
    from .utils import TestRunner
    from .things import TestThing
except ImportError:
    from test_11_rpc_e2e import TestRPCEndToEnd, TestRPCEndToEndAsync
    from utils import TestRunner
    from things import TestThing


class TestZMQ_TCP(TestRPCEndToEnd):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print("Test ZMQ TCP End to End")

    @classmethod
    def setUpThing(cls):
        """Set up the thing for the zmq object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_zmq_server(forked=True, access_points="tcp://*:5557")
        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.zmq(
                cls.server_id,
                cls.thing_id,
                "tcp://localhost:5557",
                log_level=logging.ERROR + 10,
                ignore_TD_errors=True,
            )
            return cls._client


class TestZMQAsync_TCP(TestRPCEndToEndAsync):
    @classmethod
    def setUpThing(cls):
        """Set up the thing for the zmq object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_zmq_server(forked=True, access_points="tcp://*:6000")
        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.zmq(
                cls.server_id,
                cls.thing_id,
                "tcp://localhost:6000",
                log_level=logging.ERROR + 10,
                ignore_TD_errors=True,
            )
            return cls._client


class TestZMQ_INPROC(TestRPCEndToEnd):
    @classmethod
    def setUpThing(cls):
        """Set up the thing for the zmq object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_zmq_server(forked=True, access_points="inproc")
        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.zmq(
                cls.server_id,
                cls.thing_id,
                "inproc",
                log_level=logging.ERROR + 10,
                ignore_TD_errors=True,
            )
            return cls._client


class TestZMQAsync_INPROC(TestRPCEndToEndAsync):
    @classmethod
    def setUpThing(cls):
        """Set up the thing for the zmq object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_zmq_server(forked=True, access_points="inproc")
        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.zmq(
                cls.server_id,
                cls.thing_id,
                "inproc",
                log_level=logging.ERROR + 10,
                ignore_TD_errors=True,
            )
            return cls._client


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestZMQ_TCP))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestZMQAsync_TCP))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestZMQ_INPROC))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestZMQAsync_INPROC))
    return suite


if __name__ == "__main__":
    runner = TestRunner()
    runner.run(load_tests(unittest.TestLoader(), None, None))



# Order of import is reflected in this file to avoid circular imports
from .events import *  # noqa: F403
from .actions import *  # noqa: F403
from .property import *  # noqa: F403
from .thing import *  # noqa: F403
from .meta import ThingMeta as ThingMeta



import asyncio
import warnings
import logging
import socket
import ssl
import typing
import msgspec
from copy import deepcopy
from tornado import ioloop
from tornado.web import Application
from tornado.httpserver import HTTPServer as TornadoHTTP1Server
# from tornado_http2.server import Server as TornadoHTTP2Server

from ...param import Parameterized
from ...param.parameters import Integer, IPAddress, ClassSelector, Selector, TypedList
from ...constants import HTTP_METHODS
from ...utils import (
    complete_pending_tasks_in_current_loop,
    forkable,
    get_current_async_loop,
    issubklass,
    pep8_to_dashed_name,
    get_default_logger,
    run_callable_somehow,
)
from ...config import global_config
from ...core.property import Property
from ...core.actions import Action
from ...core.events import Event
from ...core.thing import Thing, ThingMeta
from ...core.zmq.brokers import MessageMappedZMQClientPool
from ...td import ActionAffordance, EventAffordance, PropertyAffordance
from ..security import Security
from ..utils import consume_broker_queue
from .handlers import (
    ActionHandler,
    LivenessProbeHandler,
    ReadinessProbeHandler,
    PropertyHandler,
    EventHandler,
    BaseHandler,
    RWMultiplePropertiesHandler,
    StopHandler,
    ThingDescriptionHandler,
    RPCHandler,
)


class RuntimeConfig(msgspec.Struct):
    """Runtime configuration for HTTP server and handlers."""

    cors: bool = False
    """Set CORS headers for the HTTP server. If set to False, CORS headers are not set.
    This is useful when the server is used in a controlled environment where CORS is not needed.
    """


class HTTPServer(Parameterized):
    """
    HTTP(s) server to expose `Thing` over HTTP protocol. Supports HTTP 1.1.
    Use `add_thing` or `add_property` or `add_action` or `add_event` methods to add things to the server.
    """

    things = TypedList(
        item_type=(str, Thing),
        default=None,
        allow_None=True,
    )  # type: typing.List[str]
    """id(s) of the things to be served by the HTTP server."""

    port = Integer(
        default=8080,
        bounds=(1, 65535),
    )  # type: int
    """the port at which the server should be run"""

    address = IPAddress(default="0.0.0.0", doc="IP address")  # type: str
    """IP address, especially to bind to all interfaces or not"""

    logger = ClassSelector(class_=logging.Logger, default=None, allow_None=True)  # type: logging.Logger
    """logging.Logger instance"""

    log_level = Selector(
        objects=[
            logging.DEBUG,
            logging.INFO,
            logging.ERROR,
            logging.WARN,
            logging.CRITICAL,
            logging.ERROR,
        ],
        default=logging.INFO,
    )  # type: int
    """
    alternative to logger, this creates an internal logger with the specified log level 
    along with a IO stream handler.
    """

    ssl_context = ClassSelector(
        class_=ssl.SSLContext,
        default=None,
        allow_None=True,
    )  # type: typing.Optional[ssl.SSLContext]
    """SSL context to provide encrypted communication"""

    allowed_clients = TypedList(item_type=str)
    """
    Serves request and sets CORS only from these clients, other clients are rejected with 401. 
    Unlike pure CORS, the server resource is not even executed if the client is not 
    an allowed client. if None, any client is served. Not inherently a safety feature in public networks, 
    and more useful in private networks when the remote origin is known.
    """

    property_handler = ClassSelector(
        default=PropertyHandler,
        class_=(PropertyHandler, RPCHandler),
        isinstance=False,
    )  # type: typing.Union[RPCHandler, PropertyHandler]
    """custom web request handler for property read-write"""

    action_handler = ClassSelector(
        default=ActionHandler,
        class_=(ActionHandler, RPCHandler),
        isinstance=False,
    )  # type: typing.Union[RPCHandler, ActionHandler]
    """custom web request handler for actions"""

    event_handler = ClassSelector(
        default=EventHandler,
        class_=(EventHandler, RPCHandler),
        isinstance=False,
    )  # type: typing.Union[RPCHandler, EventHandler]
    """custom event handler for sending HTTP SSE"""

    security_schemes = TypedList(
        default=None,
        allow_None=True,
        item_type=Security,
    )  # type: typing.Optional[typing.List[Security]]
    """
    List of security schemes to be used by the server, 
    it is sufficient that one scheme passes for a request to be authorized.
    """

    config = ClassSelector(
        class_=RuntimeConfig,
        default=None,
        allow_None=True,
    )  # type: RuntimeConfig
    """
    Set CORS headers for the HTTP server. If set to False, CORS headers are not set. 
    This is useful when the server is used in a controlled environment where CORS is not needed
    """

    def __init__(
        self,
        *,
        port: int = 8080,
        address: str = "0.0.0.0",
        # host: typing.Optional[str] = None,
        logger: typing.Optional[logging.Logger] = None,
        log_level: int = logging.INFO,
        ssl_context: typing.Optional[ssl.SSLContext] = None,
        security_schemes: typing.Optional[typing.List[Security]] = None,
        # protocol_version : int = 1, network_interface : str = 'Ethernet',
        allowed_clients: typing.Optional[typing.Union[str, typing.Iterable[str]]] = None,
        config: typing.Optional[dict[str, typing.Any]] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        port: int, default 8080
            the port at which the server should be run
        address: str, default 0.0.0.0
            IP address
        logger: logging.Logger, optional
            logging.Logger instance
        log_level: int
            alternative to logger, this creates an internal logger with the specified log level along with a IO stream handler.
        ssl_context: ssl.SSLContext
            SSL context to provide encrypted communication
        allowed_clients: List[str]
            serves request and sets CORS only from these clients, other clients are reject with 403. Unlike pure CORS
            feature, the server resource is not even executed if the client is not an allowed client.
        **kwargs:
            additional keyword arguments for server configuration. Usually:

            - `property_handler`: `BaseHandler` | `PropertyHandler`, optional.
                custom web request handler for property read-write
            - `action_handler`: `BaseHandler` | `ActionHandler`, optional.
                custom web request handler for action
            - `event_handler`: `EventHandler` | `BaseHandler`, optional.
                custom event handler for sending HTTP SSE
        """
        config = RuntimeConfig(**(config or dict(cors=global_config.ALLOW_CORS)))
        # need to be extended when more options are added
        super().__init__(
            port=port,
            address=address,
            logger=logger,
            log_level=log_level,
            security_schemes=security_schemes,
            ssl_context=ssl_context,
            property_handler=kwargs.get("property_handler", PropertyHandler),
            action_handler=kwargs.get("action_handler", ActionHandler),
            event_handler=kwargs.get("event_handler", EventHandler),
            allowed_clients=allowed_clients if allowed_clients is not None else [],
            config=config or dict(),
        )

        self._IP = f"{self.address}:{self.port}"
        if self.logger is None:
            self.logger = get_default_logger(
                "{}|{}".format(self.__class__.__name__, f"{self.address}:{self.port}"),
                self.log_level,
            )

        self.tornado_instance = None
        self.app = Application(
            handlers=[
                (r"/stop", StopHandler, dict(owner_inst=self)),
                (r"/liveness", LivenessProbeHandler, dict(owner_inst=self)),
                (r"/readiness", ReadinessProbeHandler, dict(owner_inst=self)),
            ]
        )
        self.router = ApplicationRouter(self.app, self)

        self.zmq_client_pool = MessageMappedZMQClientPool(
            id=self._IP,
            server_ids=[],
            client_ids=[],
            handshake=False,
            poll_timeout=100,
            logger=self.logger,
        )
        self._disconnected_things = list()  # type: list[tuple[str, str, str]]

    @property
    def all_ok(self) -> bool:
        """check if all the requirements are met before starting the server, auto invoked by listen()"""
        # Add only those code here that needs to be redone always before restarting the server.
        # One time creation attributes/activities must be in init
        ioloop.IOLoop.clear_current()  # cleat the event in case any pending tasks exist, also restarting with same
        # event loop is buggy, so we remove it.
        event_loop = get_current_async_loop()  # sets async loop for a non-possessing thread as well
        # event_loop.call_soon(lambda : asyncio.create_task(self.update_router_with_things()))
        event_loop.call_soon(lambda: asyncio.create_task(self.zmq_client_pool.poll_responses()))
        # self.zmq_client_pool.handshake(), NOTE - handshake better done upfront as we already poll_responses here
        # which will prevent handshake function to succeed (although handshake will be done)

        self.tornado_event_loop = ioloop.IOLoop.current()
        # set value based on what event loop we use, there is some difference
        # between the asyncio event loop and the tornado event loop

        self.tornado_instance = TornadoHTTP1Server(self.app, ssl_options=self.ssl_context)  # type: TornadoHTTP1Server
        return True

    @forkable
    def listen(self, forked: bool = False) -> None:
        """
        Start the HTTP server. This method is blocking.
        Async event loops intending to schedule the HTTP server should instead use
        the inner tornado instance's (`HTTPServer.tornado_instance`) listen() method.

        Parameters
        ----------
        forked: bool, default `False`
            if `True`, the server is started in a separate thread and the method returns immediately.
            If `False`, the method blocks until the server is stopped.
        """
        assert self.all_ok, (
            "HTTPServer all is not ok before starting"
        )  # Will always be True or cause some other exception
        self.tornado_instance.listen(port=self.port, address=self.address)
        self.logger.info(f"started webserver at {self._IP}, ready to receive requests.")
        self.tornado_event_loop.start()
        if forked:
            complete_pending_tasks_in_current_loop()  # will reach here only when the server is stopped, so complete pending tasks

    def stop(self, attempt_async_stop: bool = True) -> None:
        """
        Stop the HTTP server - unreliable, use `async_stop()` if possible.
        A stop handler at the path `/stop` with POST method is already implemented that invokes this
        method for the clients.

        Parameters
        ----------
        attempt_async_stop: bool, default `True`
            if `True`, attempts to run the `async_stop` method to close all connections gracefully.
        """
        if attempt_async_stop:
            run_callable_somehow(self.async_stop())
            return
        self.zmq_client_pool.stop_polling()
        if not self.tornado_instance:
            return
        self.tornado_instance.stop()
        run_callable_somehow(self.tornado_instance.close_all_connections())
        if self.tornado_event_loop is not None:
            self.tornado_event_loop.stop()
        complete_pending_tasks_in_current_loop()

    async def async_stop(self) -> None:
        """
        Stop the HTTP server. A stop handler at the path `/stop` with POST method is already implemented that invokes this
        method for the clients.
        """
        self.zmq_client_pool.stop_polling()
        if not self.tornado_instance:
            return
        self.tornado_instance.stop()
        await self.tornado_instance.close_all_connections()
        if self.tornado_event_loop is not None:
            self.tornado_event_loop.stop()
            self.logger.info(f"stopped tornado event loop for server {self._IP}")

    def add_things(self, *things: Thing | dict | str) -> None:
        """
        Add things to be served by the HTTP server

        Parameters
        ----------
        *things: Thing | dict | str
            the thing instance(s) or thing classe(s) to be served, or a map of address/ZMQ protocol to thing id,
            for example - `{'tcp://my-pc:5555': 'my-thing-id', 'IPC' : 'my-thing-id-2'}`
            The server ID needs to match the thing ID, otherwise this method would be unable to find the thing.
        """
        for thing in things:
            if isinstance(thing, Thing):
                self.router.add_thing_instance(thing)
            elif isinstance(thing, ThingMeta):
                raise ValueError(
                    f"class {thing} is not a thing instance, no need to add it to the server."
                    + " Just supply a thing instance to the server. skipping..."
                )
            elif isinstance(thing, (dict, str)):
                if isinstance(thing, str):
                    # if protocol not given, try INPROC first
                    self.router.add_thing_instance_through_broker(thing, thing, access_point="INPROC")
                    if not self.zmq_client_pool.get_client_id_from_thing_id(thing):
                        self.router.add_thing_instance_through_broker(thing, thing, access_point="IPC")
                else:
                    for address, thing_id in thing.items():
                        self.router.add_thing_instance_through_broker(
                            server_id=thing_id, thing_id=thing_id, access_point=address
                        )
            elif issubklass(thing, ThingMeta):
                raise TypeError("thing should be of type Thing, not ThingMeta")
            else:
                raise TypeError(f"thing should be of type Thing, unknown type given - {type(thing)}")

    def add_thing(self, server_id: str, thing_id: str, access_point: str) -> None:
        """
        Add thing to be served by the HTTP server

        Parameters
        ----------
        thing: str | Thing
            id of the thing or the thing instance or thing class to be served
        """
        self.router.add_zmq_thing_instance(
            server_id=server_id,
            thing_id=thing_id,
            access_point=access_point,
        )

    def add_property(
        self,
        URL_path: str,
        property: Property | PropertyAffordance,
        http_methods: typing.Tuple[str, typing.Optional[str], typing.Optional[str]] | None = ("GET", "PUT", None),
        handler: BaseHandler | PropertyHandler = PropertyHandler,
        **kwargs,
    ) -> None:
        """
        Add a property to be accessible by HTTP

        Parameters
        ----------
        URL_path: str
            URL path to access the property
        property: Property | PropertyAffordance
            Property (object) to be served or its JSON representation
        http_methods: Tuple[str, str, str]
            tuple of http methods to be used for read, write and delete. Use None or omit HTTP method for
            unsupported operations. For example - for readonly property use ('GET', None, None) or ('GET',)
        handler: BaseHandler | PropertyHandler, optional
            custom handler for the property, otherwise the default handler will be used
        kwargs: dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(property, (Property, PropertyAffordance)):
            raise TypeError(f"property should be of type Property, given type {type(property)}")
        if not issubklass(handler, BaseHandler):
            raise TypeError(f"handler should be subclass of BaseHandler, given type {type(handler)}")
        if isinstance(property, Property):
            property = property.to_affordance()
        read_http_method = write_http_method = delete_http_method = None
        http_methods = self.router.adapt_http_methods(http_methods)
        if len(http_methods) == 1:
            read_http_method = http_methods[0]
        elif len(http_methods) == 2:
            read_http_method, write_http_method = http_methods
        elif len(http_methods) == 3:
            read_http_method, write_http_method, delete_http_method = http_methods
        if read_http_method != "GET":
            raise ValueError("read method should be GET")
        if write_http_method and write_http_method not in ["POST", "PUT"]:
            raise ValueError("write method should be POST or PUT")
        if delete_http_method and delete_http_method != "DELETE":
            raise ValueError("delete method should be DELETE")
        kwargs["resource"] = property
        kwargs["owner_inst"] = self
        kwargs["metadata"] = dict(http_methods=http_methods)
        self.router.add_rule(affordance=property, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def add_action(
        self,
        URL_path: str,
        action: Action | ActionAffordance,
        http_method: str | None = "POST",
        handler: BaseHandler | ActionHandler = ActionHandler,
        **kwargs,
    ) -> None:
        """
        Add an action to be accessible by HTTP

        Parameters
        ----------
        URL_path: str
            URL path to access the action
        action: Action | ActionAffordance
            Action (object) to be served or its JSON representation
        http_method: str
            http method to be used for the action
        handler: BaseHandler | ActionHandler, optional
            custom handler for the action
        kwargs: dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(action, (Action, ActionAffordance)):
            raise TypeError(f"Given action should be of type Action or ActionAffordance, given type {type(action)}")
        if not issubklass(handler, BaseHandler):
            raise TypeError(f"handler should be subclass of BaseHandler, given type {type(handler)}")
        http_methods = self.router.adapt_http_methods(http_method)
        if isinstance(action, Action):
            action = action.to_affordance()  # type: ActionAffordance
        kwargs["resource"] = action
        kwargs["owner_inst"] = self
        kwargs["metadata"] = dict(http_methods=http_methods)
        self.router.add_rule(affordance=action, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def add_event(
        self,
        URL_path: str,
        event: Event | EventAffordance | PropertyAffordance,
        handler: BaseHandler | EventHandler = EventHandler,
        **kwargs,
    ) -> None:
        """
        Add an event to be accessible by HTTP server; only GET method is supported for events.

        Parameters
        ----------
        URL_path: str
            URL path to access the event
        event: Event | EventAffordance
            Event (object) to be served or its JSON representation
        handler: BaseHandler | EventHandler, optional
            custom handler for the event
        kwargs: dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(event, (Event, EventAffordance)) and (
            not isinstance(event, PropertyAffordance) or not event.observable
        ):
            raise TypeError(f"event should be of type Event or EventAffordance, given type {type(event)}")
        if not issubklass(handler, BaseHandler):
            raise TypeError(f"handler should be subclass of BaseHandler, given type {type(handler)}")
        if isinstance(event, Event):
            event = event.to_affordance()
        kwargs["resource"] = event
        kwargs["owner_inst"] = self
        self.router.add_rule(affordance=event, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def __hash__(self):
        return hash(self._IP)

    def __eq__(self, other):
        if not isinstance(other, HTTPServer):
            return False
        return self._IP == other._IP

    def __str__(self):
        return f"{self.__class__.__name__}(address={self.address}, port={self.port})"


class ApplicationRouter:
    """
    Covering implementation (as in - a layer on top of it) of the application router to
    add rules to the tornado application. Not a real router, which is taken care of
    by the tornado application automatically.
    """

    def __init__(self, app: Application, server: HTTPServer) -> None:
        self.app = app
        self.server = server
        self._pending_rules = []

    # can add a single property, action or event rule
    def add_rule(
        self,
        affordance: PropertyAffordance | ActionAffordance | EventAffordance,
        URL_path: str,
        handler: typing.Type[BaseHandler],
        kwargs: dict,
    ) -> None:
        """
        Add rules to the application router. Note that this method will replace existing rules and can duplicate
        endpoints for an affordance without checks (i.e. you could technically add two different endpoints for the
        same affordance).
        """
        for rule in self.app.wildcard_router.rules:
            if rule.matcher == URL_path:
                warnings.warn(
                    f"URL path {URL_path} already exists in the router -"
                    + f" replacing it for {affordance.what} {affordance.name}",
                    category=UserWarning,
                )
        for rule in self._pending_rules:
            if rule[0] == URL_path:
                warnings.warn(
                    f"URL path {URL_path} already exists in the pending rules -"
                    + f" replacing it for {affordance.what} {affordance.name}",
                    category=UserWarning,
                )
        if getattr(affordance, "thing_id", None) is not None:
            if not URL_path.startswith(f"/{affordance.thing_id}"):
                warnings.warn(
                    f"URL path {URL_path} does not start with the thing id {affordance.thing_id},"
                    + f" adding it to the path, new path = {f'/{affordance.thing_id}{URL_path}'}. "
                    + " This warning can be usually safely ignored."
                )
                URL_path = f"/{affordance.thing_id}{URL_path}"
            self.app.wildcard_router.add_rules([(URL_path, handler, kwargs)])
        elif affordance.thing_cls is not None:
            self._pending_rules.append((URL_path, handler, kwargs))
        else:
            raise RuntimeError("object has no thing id or thing class associated with it, cannot add rule")
        """
        for handler based tornado rule matcher, the Rule object has following
        signature
        
        def __init__(
            self,
            matcher: "Matcher",
            target: Any,
            target_kwargs: Optional[Dict[str, Any]] = None,
            name: Optional[str] = None,
        ) -> None:

        matcher - based on route
        target - handler
        target_kwargs - given to handler's initialize
        name - ...

        len == 2 tuple is route + handler
        len == 3 tuple is route + handler + target kwargs
    
        so we give (path, BaseHandler, {'resource' : PropertyAffordance, 'owner' : self})
        
        path is extracted from interaction affordance name or given by the user
        BaseHandler is the base handler of this package for interaction affordances
        resource goes into target kwargs which is needed for the handler to work correctly
        """

    # can add multiple properties, actions and events at once
    def add_interaction_affordances(
        self,
        properties: typing.Iterable[PropertyAffordance],
        actions: typing.Iterable[ActionAffordance],
        events: typing.Iterable[EventAffordance],
        thing_id: str = None,
    ) -> None:
        for property in properties:
            if property in self:
                continue
            route = self.adapt_route(property.name)
            if property.thing_id is not None:
                path = f"/{property.thing_id}{route}"
            self.server.add_property(
                URL_path=path,
                property=property,
                http_methods=("GET")
                if property.readOnly
                else (
                    "GET",
                    "PUT",
                ),  # if prop.fdel is None else ('GET', 'PUT', 'DELETE'),
                handler=self.server.property_handler,
            )
            if property.observable:
                self.server.add_event(
                    URL_path=f"{path}/change-event",
                    event=property,
                    handler=self.server.event_handler,
                )
        for action in actions:
            if action in self:
                continue
            route = self.adapt_route(action.name)
            if action.thing_id is not None:
                path = f"/{action.thing_id}{route}"
            self.server.add_action(URL_path=path, action=action, handler=self.server.action_handler)
        for event in events:
            if event in self:
                continue
            route = self.adapt_route(event.name)
            if event.thing_id is not None:
                path = f"/{event.thing_id}{route}"
            self.server.add_event(URL_path=path, event=event, handler=self.server.event_handler)

        # thing description handler
        get_thing_model_action = next((action for action in actions if action.name == "get_thing_model"), None)
        get_thing_description_action = deepcopy(get_thing_model_action)
        get_thing_description_action.override_defaults(name="get_thing_description")
        self.server.add_action(
            URL_path=f"/{thing_id}/resources/wot-td" if thing_id else "/resources/wot-td",
            action=get_thing_description_action,
            http_method=("GET",),
            handler=ThingDescriptionHandler,
        )

        # RW multiple properties handler
        read_properties = Thing._get_properties.to_affordance(Thing)
        write_properties = Thing._set_properties.to_affordance(Thing)
        read_properties.override_defaults(thing_id=get_thing_model_action.thing_id)
        write_properties.override_defaults(thing_id=get_thing_model_action.thing_id)
        self.server.add_action(
            URL_path=f"/{thing_id}/properties" if thing_id else "/properties",
            action=read_properties,
            http_method=("GET", "PUT", "PATCH"),
            handler=RWMultiplePropertiesHandler,
            read_properties_resource=read_properties,
            write_properties_resource=write_properties,
        )

    # can add an entire thing instance at once
    def add_thing_instance(self, thing: Thing) -> None:
        """
        internal method to add a thing instance to be served by the HTTP server. Iterates through the
        interaction affordances and adds a route for each property, action and event.
        """
        # Prepare affordance lists with error handling (single loop)
        properties, actions, events = [], [], []
        affordances = [
            (thing.properties.remote_objects.values(), properties, "property"),
            (thing.actions.descriptors.values(), actions, "action"),
            (thing.events.descriptors.values(), events, "event"),
        ]
        for objs, target_list, affordance_type in affordances:
            for obj in objs:
                try:
                    target_list.append(obj.to_affordance(thing))
                except Exception as ex:
                    self.server.logger.error(
                        f"Failed to convert {affordance_type} {getattr(obj, 'name', obj)} to affordance: {ex}"
                    )

        self._resolve_rules(thing.id, thing.__class__)
        self.add_interaction_affordances(
            properties,
            actions,
            events,
            thing_id=thing.id,
        )

    # can add an entire thing instance at once over ZMQ
    def add_thing_instance_through_broker(self, server_id: str, thing_id: str, access_point: str) -> None:
        """
        Add a thing served by ZMQ server to the HTTP server. Mostly useful for INPROC transport which behaves like a local object.
        Iterates through the interaction affordances and adds a route for each property, action and event.
        """
        run_callable_somehow(
            self.async_add_thing_instance_through_broker(
                thing_id=thing_id, server_id=server_id, access_point=access_point
            )
        )

    async def async_add_thing_instance_through_broker(
        self,
        server_id: str,
        thing_id: str,
        access_point: str = None,
    ) -> None:
        try:
            self.server._disconnected_things.append((server_id, thing_id, access_point))
            client, TD = await consume_broker_queue(
                id=self.server._IP,
                server_id=server_id,
                thing_id=thing_id,
                access_point=access_point,
            )
            # Add routes to server
            self.add_interaction_affordances(
                [PropertyAffordance.from_TD(name, TD) for name in TD["properties"].keys()],
                [ActionAffordance.from_TD(name, TD) for name in TD["actions"].keys()],
                [EventAffordance.from_TD(name, TD) for name in TD["events"].keys()],
                thing_id=thing_id,
            )
            # add client to pool
            self.server.zmq_client_pool.register(client, thing_id)
            self.server._disconnected_things.remove((server_id, thing_id, access_point))
        except ConnectionError:
            self.server.logger.warning(
                f"could not connect to {thing_id} using on server {server_id} with access_point {access_point}"
            )
        except Exception as ex:
            self.server.logger.error(
                f"could not connect to {thing_id} using on server {server_id} with access_point {access_point}. error: {str(ex)}"
            )

    def _resolve_rules(self, thing_id: str, thing_cls: ThingMeta) -> None:
        """
        Process the pending rules and add them to the application router.
        """
        pending_rules = []
        for rule in self._pending_rules:
            affordance = rule[2].get("resource", None)  # type: PropertyAffordance | ActionAffordance | EventAffordance
            if not affordance or affordance.owner != thing_cls:
                pending_rules.append(rule)
                continue
            URL_path, handler, kwargs = rule
            URL_path = f"/{thing_id}{URL_path}"
            rule = (URL_path, handler, kwargs)
            self.app.wildcard_router.add_rules([rule])
        self._pending_rules = pending_rules

    def __contains__(
        self,
        item: str | Property | Action | Event | PropertyAffordance | ActionAffordance | EventAffordance,
    ) -> bool:
        """
        Check if the item is in the application router.
        Not exact for torando's rules when a string is provided for the URL path,
        as you need to provide the Matcher object
        """
        if isinstance(item, str):
            for rule in self.app.wildcard_router.rules:
                if rule.matcher == item:
                    return True
                if hasattr(rule.matcher, "regex") and rule.matcher.regex.match(item):
                    return True
            for rule in self._pending_rules:
                if rule[0] == item:
                    return True
        elif isinstance(item, (Property, Action, Event)):
            item = item.to_affordance()
        if isinstance(item, (PropertyAffordance, ActionAffordance, EventAffordance)):
            for rule in self.app.wildcard_router.rules:
                if rule.target_kwargs.get("resource", None) == item:
                    return True
            for rule in self._pending_rules:
                if rule[2].get("resource", None) == item:
                    return True
        return False

    def get_href_for_affordance(self, affordance, authority: str = None, use_localhost: bool = False) -> str:
        if affordance not in self:
            raise ValueError(f"affordance {affordance} not found in the application router")
        for rule in self.app.wildcard_router.rules:
            if rule.target_kwargs.get("resource", None) == affordance:
                path = str(rule.matcher.regex.pattern).rstrip("$")
                return f"{self.get_basepath(authority, use_localhost)}{path}"

    def get_basepath(self, authority: str = None, use_localhost: bool = False) -> str:
        if authority:
            return authority
        protocol = "https" if self.server.ssl_context else "http"
        port = f":{self.server.port}" if self.server.port != 80 else ""
        if not use_localhost:
            return f"{protocol}://{socket.gethostname()}{port}"
        if self.server.address == "0.0.0.0" or self.server.address == "127.0.0.1":
            return f"{protocol}://127.0.0.1{port}"
        elif self.server.address == "::":
            return f"{protocol}://[::1]{port}"
        return f"{protocol}://localhost{port}"

    basepath = property(fget=get_basepath, doc="basepath of the server")

    def get_target_kwargs_for_affordance(self, affordance) -> dict:
        """
        Get the target kwargs for the affordance in the application router.
        """
        if affordance not in self:
            raise ValueError(f"affordance {affordance} not found in the application router")
        for rule in self.app.wildcard_router.rules:
            if rule.target_kwargs.get("resource", None) == affordance:
                return rule.target_kwargs
        for rule in self._pending_rules:
            if rule[2].get("resource", None) == affordance:
                return rule[2]
        raise ValueError(f"affordance {affordance} not found in the application router rules")

    def adapt_route(self, interaction_affordance_name: str) -> str:
        if interaction_affordance_name == "get_thing_model":
            return "/resources/wot-tm"
        return f"/{pep8_to_dashed_name(interaction_affordance_name)}"

    def adapt_http_methods(self, http_methods: typing.Any):
        """comply the supplied HTTP method to the router to a tuple and check if the method is supported"""
        if isinstance(http_methods, str):
            http_methods = (http_methods,)
        if not isinstance(http_methods, tuple):
            raise TypeError("http_method should be a tuple")
        for method in http_methods:
            if method not in HTTP_METHODS.__members__.values() and method is not None:
                raise ValueError(f"method {method} not supported")
        return http_methods

    def print_rules(self) -> None:
        """
        Print the rules in the application router.
        prettytable is used if available, otherwise a simple print is done.
        """
        try:
            from prettytable import PrettyTable

            table = PrettyTable()
            table.field_names = ["URL Path", "Handler", "Resource Name"]

            for rule in self.app.wildcard_router.rules:
                table.add_row(
                    [
                        rule.matcher,
                        rule.target.__name__,
                        getattr(rule.target_kwargs.get("resource"), "name", "N/A"),
                    ]
                )
            for rule in self._pending_rules:
                table.add_row([rule[0], rule[1].__name__, rule[2]["resource"].name])
            print(table)
        except ImportError:
            print("Application Router Rules:")
            for rule in self.app.wildcard_router.rules:
                print(rule)
            for rule in self._pending_rules:
                print(rule[0], rule[2]["resource"].name)


__all__ = [HTTPServer.__name__]



import typing
from enum import Enum

from ..param.parameterized import Parameter, Parameterized, ParameterizedMetaclass
from ..utils import issubklass
from ..exceptions import StateMachineError
from ..schema_validators import JSONSchemaValidator
from .dataklasses import RemoteResourceInfoValidator
from .events import Event, EventDispatcher  # noqa: F401


class Property(Parameter):
    """
    get/set/delete an object/instance attribute with type definitions, validations, post-set effects, metadata and more.
    Please note the capital 'P' in `Property` to differentiate from python's own `property`.
    `Property` objects are similar to python's `property` but not a subclass of it due to limitations and redundancy.
    """

    __slots__ = [
        "db_persist",
        "db_init",
        "db_commit",
        "model",
        "metadata",
        "fcomparator",
        "validator",
        "execution_info",
        "_execution_info_validator",
        "_observable_event_descriptor",
        "_old_value_internal_name",
    ]

    def __init__(
        self,
        default: typing.Any = None,
        *,
        doc: typing.Optional[str] = None,
        constant: bool = False,
        readonly: bool = False,
        allow_None: bool = False,
        label: typing.Optional[str] = None,
        state: typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,
        db_persist: bool = False,
        db_init: bool = False,
        db_commit: bool = False,
        observable: bool = False,
        model: typing.Optional["BaseModel"] = None,
        class_member: bool = False,
        fget: typing.Optional[typing.Callable] = None,
        fset: typing.Optional[typing.Callable] = None,
        fdel: typing.Optional[typing.Callable] = None,
        fcomparator: typing.Optional[typing.Callable] = None,
        deepcopy_default: bool = False,
        per_instance_descriptor: bool = False,
        remote: bool = True,
        precedence: typing.Optional[float] = None,
        metadata: typing.Optional[typing.Dict] = None,
    ) -> None:
        """
        Parameters
        ----------

        default: None or corresponding to property type
            The default value of the property.

        doc: str, default empty
            docstring explaining what this property represents.

        allow_None: bool, default `False`
            if `True`, None is accepted as a valid value in addition to any other values that are allowed.

        readonly: bool, default `False`
            if `True`, the value cannot be changed at all. Either the value is fetched always or a getter method
            is executed which may still generate dynamic values at each read/get operation.

        constant: bool, default `False`
            if `True`, the value can be changed only once when `allow_None` is set to `True`. The
            value is otherwise constant on the `Thing` instance.

        observable: bool, default `False`
            if `True`, pushes change events after every read and write if the value is different from the previous operation.
            Supply a function if interested to evaluate on what conditions the change event must be emitted.
            Default condition is a plain not-equal-to operator.

        state: str | Enum, default None
            state of state machine where property write can be executed

        db_persist: bool, default `False`
            if `True`, every write is stored in a database and the property value persists `Thing` instance destruction and
            recreation. The loaded value from database is written into the property at `Thing.__post_init__`.
            set `Thing.use_default_db` to `True`, to set up a default SQLite database or supply a `db_config_file`.

        db_init: bool, default `False`
            if `True`, property's first value is loaded from database and written using setter.
            Further writes are not written to database. Use this to load predetermined configuration which may change at
            a later time temporarily. if `db_persist` is `True`, this value is ignored.

        db_commit: bool,
            if `True`, all write values are stored to database. The database value is not loaded at `Thing.__post_init__()`.
            if db_persist is `True`, this value is ignored.

        fget: Callable, default None
            custom getter method, mandatory when setter method is also custom supplied.

        fset: Callable, default None
            custom setter method

        fdel: Callable, default None
            custom deleter method

        remote: bool, default `True`
            set to `False` to make the property local/not remotely accessible

        label: str, default extracted from object name
            optional text label to be used when this Property is shown in a listing. If no label is supplied,
            the attribute name for this property in the owning `Thing` object is used.

        metadata: dict, default None
            store your own metadata for the property which gives useful (and modifiable) information
            about the property. Properties operate using slots which means you cannot set foreign attributes on this object
            normally. This metadata dictionary should overcome this limitation.

        per_instance_descriptor: bool, default `False`
            whether a separate Property instance will be created for every `Thing` instance. `False` by default.
            If `False`, all instances of a `Thing` class will share the same Property object, including all validation
            attributes (bounds, allow_None etc.).

        deepcopy_default: bool, default `False`
            controls whether the default value of this Property will be deepcopied when a `Thing` object is instantiated (if
            `True`), or if the single default value will be shared by all `Thing` instances (if `False`). For an immutable
            Property value, it is best to leave deep_copy at the default of `False`. For a mutable Property value,
            for example - lists and dictionaries, the default of `False` is also appropriate if you want all instances to share
            the same value state, e.g. if they are each simply referring to a single global object like a singleton.
            If instead each `Thing` should have its own independently mutable value, deep_copy should be set to
            `True`. This setting is similar to using `field`'s `default_factory` in python dataclasses.

        class_member : bool, default `False`
            when `True`, property is set on `Thing` class instead of `Thing` instance.

        precedence: float, default None
            a numeric value, usually in the range 0.0 to 1.0, which allows the order of Properties in a class to be defined in
            a listing or e.g. in GUI menus. A negative precedence indicates a property that should be hidden in such listings.

        """
        super().__init__(
            default=default,
            doc=doc,
            constant=constant,
            readonly=readonly,
            allow_None=allow_None,
            label=label,
            per_instance_descriptor=per_instance_descriptor,
            deepcopy_default=deepcopy_default,
            class_member=class_member,
            fget=fget,
            fset=fset,
            fdel=fdel,
            precedence=precedence,
        )
        self.db_persist = db_persist
        self.db_init = db_init
        self.db_commit = db_commit
        self.fcomparator = fcomparator
        self.metadata = metadata
        self._observable_event_descriptor = None
        if observable:
            self._observable_event_descriptor = Event()
        self._execution_info_validator = None
        self.execution_info = None  # typing.Optional[RemoteResource]
        if remote:
            self._execution_info_validator = RemoteResourceInfoValidator(state=state, isproperty=True, obj=self)
            self.execution_info = self._execution_info_validator  # TODO: use dataclass or remove this attribute
        self.model = None
        self.validator = None
        if model:
            if isinstance(model, dict):
                self.model = model
                self.validator = JSONSchemaValidator(model).validate
            else:
                self.model = wrap_plain_types_in_rootmodel(model)  # type: BaseModel
                self.validator = self.model.model_validate

    def __set_name__(self, owner: typing.Any, attrib_name: str) -> None:
        super().__set_name__(owner, attrib_name)
        if self._execution_info_validator:
            self._execution_info_validator.obj_name = attrib_name
        if self._observable_event_descriptor:
            _observable_event_name = f"{self.name}_change_event"
            self._old_value_internal_name = f"{self._internal_name}_old_value"
            self._observable_event_descriptor.doc = f"change event for {self.name}"
            self._observable_event_descriptor._observable = True
            self._observable_event_descriptor.__set_name__(owner, _observable_event_name)
            # This is a descriptor object, so we need to set it on the owner class
            setattr(owner, _observable_event_name, self._observable_event_descriptor)

    def __get__(self, obj: Parameterized, objtype: ParameterizedMetaclass) -> typing.Any:
        read_value = super().__get__(obj, objtype)
        self.push_change_event(obj, read_value)
        return read_value

    def push_change_event(self, obj, value: typing.Any) -> None:
        """
        Pushes change event both on read and write if an event publisher object is available
        on the owning `Thing`.

        Parameters
        ----------
        obj: Thing
            the `Thing` instance owning this property
        value: Any
            the value that was just read or written
        """
        if obj is None:
            return
        if self._observable_event_descriptor and obj.event_publisher:
            event_dispatcher = getattr(obj, self._observable_event_descriptor.name, None)  # type: EventDispatcher
            old_value = obj.__dict__.get(self._old_value_internal_name, NotImplemented)
            obj.__dict__[self._old_value_internal_name] = value
            if self.fcomparator:
                if issubklass(self.fcomparator, classmethod):
                    if not self.fcomparator(self.owner, old_value, value):
                        return
                elif not self.fcomparator(obj, old_value, value):
                    return
            elif not old_value != value:
                return
            event_dispatcher.push(value)

    def validate_and_adapt(self, value) -> typing.Any:
        """
        Validate the given value and adapt it if a proper logical reasoning can be given,
        for example, cropping a number to its bounds. Returns modified value.
        """
        if value is None:
            if self.allow_None:
                return
            else:
                raise ValueError(f"Property {self.name} does not allow None values")
        if self.model:
            if isinstance(self.model, dict):
                self.validator(value)
            elif issubklass(self.model, RootModel):
                value = self.model(value)
            elif issubklass(self.model, BaseModel):
                value = self.model(**value)
        return super().validate_and_adapt(value)

    def external_set(self, obj: Parameterized, value: typing.Any) -> None:
        """
        method called when the value of the property is set from an external source, e.g. a remote client.
        Usually introduces a state machine check before allowing the set operation.
        """
        if self.execution_info.state is None or (
            hasattr(obj, "state_machine") and obj.state_machine.current_state in self.execution_info.state
        ):
            return self.__set__(obj, value)
        else:
            raise StateMachineError(
                "Thing {} is in `{}` state, however attribute can be written only in `{}` state".format(
                    obj.id, obj.state_machine.current_state, self.execution_info.state
                )
            )

    def _post_value_set(self, obj, value: typing.Any) -> None:
        if (self.db_persist or self.db_commit) and hasattr(obj, "db_engine"):
            from .thing import Thing

            assert isinstance(obj, Thing), (
                f"database property {self.name} bound to a non Thing, currently not supported"
            )
            obj.db_engine.set_property(self, value)
        self.push_change_event(obj, value)
        return super()._post_value_set(obj, value)

    def comparator(self, func: typing.Callable) -> typing.Callable:
        """
        Register a comparator method using this decorator to decide when to push
        a change event.

        Signature of the comparator method must be:
        ```
        def func(self, old_value, new_value) -> bool
        ```

        Parameters
        ----------
        func: Callable
            a method which takes two arguments - the old value and new value - and returns
            `True` if the values are considered different enough to push a change event.
            If the method is a classmethod, the first argument will be the owning class.
            If the method is a normal method, the first argument will be the owning instance.
            The second and third arguments will be the old and new values respectively.

        Returns
        -------
        Callable
            the supplied function, no modifications made but a reference is stored
        """
        self.fcomparator = func
        return func

    @property
    def is_remote(self):
        """`False` if the property is not remotely accessible"""
        return self._execution_info_validator is not None

    @property
    def observable(self) -> bool:
        """`True` if the property pushes change events on read and write"""
        return self._observable_event_descriptor is not None

    def to_affordance(self, owner_inst=None):
        """
        Generates a `PropertyAffordance` TD fragment for this Property

        Parameters
        ----------
        owner_inst: Thing, optional
            The instance of the owning `Thing` object. If not supplied, the class is used.

        Returns
        -------
        PropertyAffordance
            the affordance TD fragment for this property
        """
        from ..td import PropertyAffordance

        return PropertyAffordance.generate(self, owner_inst or self.owner)


try:
    from pydantic import BaseModel, RootModel, create_model, ConfigDict

    def wrap_plain_types_in_rootmodel(model: type) -> type[BaseModel] | type[RootModel]:
        """
        Ensure a type is a subclass of BaseModel.

        If a `BaseModel` subclass is passed to this function, we will pass it
        through unchanged. Otherwise, we wrap the type in a RootModel.
        In the future, we may explicitly check that the argument is a type
        and not a model instance.
        """
        if model is None:
            return
        if issubklass(model, BaseModel):
            return model
        return create_model(
            f"{model!r}",
            root=(model, ...),
            __base__=RootModel,
            __config__=ConfigDict(arbitrary_types_allowed=True),
        )  # type: ignore[call-overload]
except ImportError:

    def wrap_plain_types_in_rootmodel(model: type) -> type:
        raise ImportError("pydantic is not installed, please install it to use this feature") from None


__all__ = [Property.__name__]



def copy_parameters(
    src: str = "D:/onedrive/desktop/dashboard/scada/scadapy/scadapy/param/parameters.py",
    dst: str = "D:/onedrive/desktop/dashboard/scada/scadapy/scadapy/server/remote_parameters.py",
) -> None:
    skip_classes = [
        "Infinity",
        "resolve_path",
        "normalize_path",
        "BaseConstrainedList",
        "TypeConstrainedList",
        "TypeConstrainedDict",
        "TypedKeyMappingsConstrainedDict",
        "Event",
    ]
    end_line = "def hashable"
    additional_imports = [
        "from ..param.parameters import (TypeConstrainedList, TypeConstrainedDict, abbreviate_paths,\n",
        "                       TypedKeyMappingsConstrainedDict, resolve_path, concrete_descendents, named_objs)\n"
        "from .remote_parameter import RemoteParameter\n",
        "from .constants import HTTP, PROXY, USE_OBJECT_NAME, GET, PUT",
    ]

    def fetch_line() -> typing.Generator[str]:
        with open(src, "r") as file:
            oldlines = file.readlines()
            for line in oldlines:
                yield line

    remote_init_kwargs = [
        "\t\t\tURL_path : str = USE_OBJECT_NAME, http_method : typing.Tuple[str, str] = (GET, PUT),\n",
        "\t\t\tstate : typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,\n",
        "\t\t\tdb_persist : bool = False, db_init : bool = False, db_commit : bool = False,\n"
        "\t\t\taccess_type : str = (HTTP, PROXY),\n",
    ]

    remote_super_init = [
        "\t\t\tURL_path=URL_path, http_method=http_method, state=state, db_persist=db_persist,\n",
        "\t\t\tdb_init=db_init, db_commit=db_commit, access_type=access_type)\n",
    ]

    common_linegen = fetch_line()
    newlines = []

    def skip_to_init_doc():
        for line in common_linegen:
            if "doc : typing.Optional[str] = None" in line:
                return line
            else:
                newlines.append(line)

    def skip_to_super_init_end():
        for line in common_linegen:
            if "precedence=precedence)" in line:
                return line
            else:
                newlines.append(line)

    def is_function(line: str) -> bool:
        if "def " in line and "self" not in line and "cls" not in line and "obj" not in line:
            return True
        return False

    def next_line_after_skip_class_or_function() -> str:
        for line_ in common_linegen:
            if ("class " in line_ and ":" in line_) or is_function(line_):
                return line_

    def process_current_line(line: str):
        newline = line
        if "import " in line and "parameterized " in line:
            newlines_ = [
                line.replace("from .parameterized", "from ..param.parameterized").replace("ParamOverrides,", ""),
                next(common_linegen).replace("ParameterizedFunction, descendents,", ""),
                *additional_imports,
            ]
            newlines.extend(newlines_)
            return
        elif "from collections import OrderedDict" in line:
            newlines.append("from enum import Enum\n")
        elif "from .utils" in line or "from .exceptions" in line:
            newline = line.replace("from .", "from ..param.")
        elif "class " in line and ":" in line and line.startswith("class"):
            if "(Parameter):" in line:
                newline = line.replace("(Parameter):", "(RemoteParameter):")
                newlines.append(newline)
            else:
                classname_with_inheritance = line.split(" ", 1)[1][:-2]  # [:-2] for removing colon
                classname_without_inheritance = classname_with_inheritance.split("(", 1)[0]
                if classname_without_inheritance in skip_classes:
                    newline = next_line_after_skip_class_or_function()
                    process_current_line(newline)
                    return
                else:
                    newlines.append(line)
            newline = skip_to_init_doc()
            newlines.append(newline)
            newlines.extend(remote_init_kwargs)
            newline = skip_to_super_init_end()
            if newline:
                newline = newline.replace("precedence=precedence)", "precedence=precedence,")
                newlines.append(newline)
                newlines.extend(remote_super_init)
            return
        elif "Parameter.__init__" in line:
            newline = line.replace("Parameter.__init__", "RemoteParameter.__init__")
        elif is_function(line):
            newline = next_line_after_skip_class_or_function()
            process_current_line(newline)
            return
        newlines.append(newline)

    for line in common_linegen:
        process_current_line(line)
        if end_line in line:
            newlines.pop()
            break

    with open(dst, "w") as file:
        file.writelines(newlines)



from typing import Any, Optional
import uuid
import zmq.asyncio
import logging

from ..config import global_config
from ..constants import Operations
from ..core import Thing, Action
from ..core.zmq import AsyncZMQClient, AsyncEventConsumer
from ..td.interaction_affordance import EventAffordance


async def consume_broker_queue(
    id: str,
    server_id: str,
    thing_id: str,
    access_point: str,
    context: Optional[zmq.asyncio.Context] = None,
    logger: Optional[logging.Logger] = None,
    poll_timeout: int = 1000,
) -> tuple[AsyncZMQClient, dict[str, Any]]:
    """
    Connect to a running Thing via ZMQ INPROC and fetch its Thing Description.

    Parameters
    ----------
    id : str
        Unique identifier for the client.
    server_id : str
        The server ID to connect to.
    thing_id : str
        The Thing ID whose Thing Description (TD) is to be fetched.
    access_point : str
        The access point (e.g., "TCP", "WS", or a specific address).
    context : Optional[zmq.asyncio.Context], optional
        ZMQ context to use for the connection. If None, uses the global context.
    logger : Optional[logging.Logger], optional
        Logger instance for logging events. If None, no logging is performed.
    poll_timeout : int, optional
        Poll timeout in milliseconds (default is 1000).

    Returns
    -------
    tuple[AsyncZMQClient, dict[str, Any]]
        A tuple containing the connected AsyncZMQClient and the fetched Thing Description as a dictionary.
    """
    from ..client.zmq.consumed_interactions import ZMQAction

    # create client
    client = AsyncZMQClient(
        id=id,
        server_id=server_id,
        access_point=access_point,
        context=context or global_config.zmq_context(),
        handshake=False,
        logger=logger,
        poll_timeout=poll_timeout,
    )
    # connect client
    client.handshake(10000)
    await client.handshake_complete(10000)

    # fetch ZMQ INPROC TD
    assert isinstance(Thing.get_thing_model, Action)  # type definition
    FetchTMAffordance = Thing.get_thing_model.to_affordance()
    FetchTMAffordance.override_defaults(thing_id=thing_id, name="get_thing_description")
    fetch_td = ZMQAction(
        resource=FetchTMAffordance,
        sync_client=None,
        async_client=client,
        logger=logger,
        owner_inst=None,
    )
    if isinstance(access_point, str) and len(access_point) in [3, 6]:
        access_point = access_point.upper()
    elif access_point.lower().startswith("tcp://"):
        access_point = "TCP"
    TD = await fetch_td.async_call(ignore_errors=True, protocol=access_point)  # type: dict[str, Any]
    return client, TD


def consume_broker_pubsub_per_event(resource: EventAffordance) -> AsyncEventConsumer:
    if isinstance(resource, EventAffordance):
        form = resource.retrieve_form(Operations.subscribeevent)
    else:
        form = resource.retrieve_form(Operations.observeproperty)
    return AsyncEventConsumer(
        id=f"{resource.name}|EventTunnel|{uuid.uuid4().hex[:8]}",
        event_unique_identifier=f"{resource.thing_id}/{resource.name}",
        access_point=form.href,
        context=global_config.zmq_context(),
        logger=global_config.logger(),
    )



"""
MIT License

Copyright (c) 2018 CTIC Centro Tecnologico

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# copied from wotpy repository

import asyncio
import threading
import typing
import builtins
import logging
from dataclasses import dataclass

from ..td import PropertyAffordance, ActionAffordance, EventAffordance
from ..td.forms import Form
from ..utils import get_current_async_loop
from ..constants import Operations


class ConsumedThingAction:
    # Client side action call abstraction. Subclasss from here to implement protocol specific action call.
    # Dont add class doc otherwise __doc__ in slots will conflict with class variable

    def __init__(
        self,
        resource: ActionAffordance,
        owner_inst: typing.Any,
        logger: logging.Logger,
        # schema_validator: typing.Type[BaseSchemaValidator] | None = None
    ) -> None:
        """
        Parameters
        ----------
        resource: ActionAffordance
            dataclass TD fragment representing the action (must have forms).
        owner_inst: typing.Optional[typing.Any]
            instance of the owning consumed Thing or `ObjectProxy`
        logger: logging.Logger
            logger instance
        """
        self.resource = resource
        self.owner_inst = owner_inst
        self.logger = logger
        self.schema_validator = None  # schema_validator

    def get_last_return_value(self, raise_exception: bool = False) -> typing.Any:
        """retrieve return value of the last call to the action"""
        raise NotImplementedError("implement get_last_return_value per protocol")

    last_return_value = property(
        fget=get_last_return_value,
        doc="cached return value of the last call to the method",
    )
    """cached return value of the last call to the method"""

    def __call__(self, *args, **kwargs) -> typing.Any:
        """
        Invoke action/method on server

        Parameters
        ----------
        *args: typing.Any
            arguments to the action
        **kwargs: typing.Any
            keyword arguments to the action

        Returns
        -------
        typing.Any
            reply of the action call
        """
        raise NotImplementedError("implement action __call__ per protocol")

    async def async_call(self, *args, **kwargs) -> typing.Any:
        """
        async invoke action on server - asynchronous at the network level, may not necessarily be at the server level.

        Parameters
        ----------
        *args: typing.Any
            arguments to the action
        **kwargs: typing.Any
            keyword arguments to the action

        Returns
        -------
        typing.Any
            reply of the action call
        """
        raise NotImplementedError("implement action async_call per protocol")

    def oneway(self, *args, **kwargs) -> None:
        """
        Only invokes the action on the server and does not wait for reply,
        neither does the server reply to this invokation.

        Parameters
        ----------
        *args: typing.Any
            arguments to the action
        **kwargs: typing.Any
            keyword arguments to the action
        """
        raise NotImplementedError("implement action oneway call per protocol")

    def noblock(self, *args, **kwargs) -> str:
        """
        Invoke the action and collect the reply later. A message ID must be returned by the server to identify the
        invokation.

        Parameters
        ----------
        *args: typing.Any
            arguments to the action
        **kwargs: typing.Any
            keyword arguments to the action

        Returns
        -------
        str
            id of the request or message (UUID4 as string)
        """
        raise NotImplementedError("implement action noblock call per protocol")

    def read_reply(self, message_id: str, timeout: float | int | None = None) -> typing.Any:
        """
        Read the reply of the action call which was scheduled with `noblock`.

        Parameters
        ----------
        message_id: str
            id of the request or message (UUID4 as string)
        timeout: float | int | None
            timeout in seconds to wait for the reply, None means wait indefinitely

        Returns
        -------
        typing.Any
            reply of the action call
        """
        raise NotImplementedError("implement action read_reply per protocol")

    def __hash__(self):
        return hash(self.resource.name)

    def __eq__(self, other):
        if not isinstance(other, ConsumedThingAction):
            return False
        return self.resource.name == other.resource.name


class ConsumedThingProperty:
    # property get set abstraction
    # Dont add doc otherwise __doc__ in slots will conflict with class variable

    def __init__(
        self, resource: PropertyAffordance, owner_inst: typing.Optional[typing.Any], logger: logging.Logger
    ) -> None:
        """
        Parameters
        ----------
        resource: PropertyAffordance
            dataclass object TD fragment representing the property (must have forms).
        owner_inst: typing.Optional[typing.Any]
            instance of the owning consumed Thing or `ObjectProxy`
        logger: logging.Logger
            logger instance
        """
        self.resource = resource
        self.owner_inst = owner_inst
        self.logger = logger

    @property  # i.e. cannot have setter
    def last_read_value(self) -> typing.Any:
        """cache of last read value"""
        raise NotImplementedError("implement last_read_value per protocol")

    def set(self, value: typing.Any) -> None:
        """
        Set or write property value.

        Parameters
        ----------
        value: typing.Any
            value to set
        """
        raise NotImplementedError("implement property set per protocol")

    def get(self) -> typing.Any:
        """
        Get or read property value.

        Returns
        -------
        typing.Any
            property value
        """
        raise NotImplementedError("implement property get per protocol")

    async def async_set(self, value: typing.Any) -> None:
        """
        Async set or write property value - asynchronous at the network level, may not necessarily be at the server level.

        Parameters
        ----------
        value: typing.Any
            value to set
        """
        raise NotImplementedError("implement async property set per protocol")

    async def async_get(self) -> typing.Any:
        """
        Async get or read property value - asynchronous at the network level, may not necessarily be at the server level.

        Returns
        -------
        typing.Any
            property value
        """
        raise NotImplementedError("implement async property get per protocol")

    def noblock_get(self) -> str:
        """
        Get or read property value without blocking, i.e. make a request and collect it later and the method returns immediately.
        Server must return a message ID to identify the request.

        Returns
        -------
        str
            id of the request or message (UUID4 as string)
        """
        raise NotImplementedError("implement property noblock get per protocol")

    def noblock_set(self, value: typing.Any) -> str:
        """
        Set or write property value without blocking, i.e. make a request and collect it later and the method returns immediately.
        Server must return a message ID to identify the request.

        Parameters
        ----------
        value: typing.Any
            value to set

        Returns
        -------
        str
            id of the request or message (UUID4 as string)
        """
        raise NotImplementedError("implement property noblock set per protocol")

    def oneway_set(self, value: typing.Any) -> None:
        """
        Set property value without waiting for acknowledgement. The server also does not send any reply.
        There is no guarantee that the property value was set.

        Parameters
        ----------
        value: typing.Any
            value to set
        """
        raise NotImplementedError("implement property oneway set per protocol")

    def observe(self, *callbacks: typing.Callable) -> None:
        """
        Observe property value changes

        Parameters
        ----------
        *callbacks: typing.Callable
            callback to call when property value changes
        """
        # looks like this will be unused. observe property is done via ConsumedThingEvent
        raise NotImplementedError("implement property observe per protocol")

    def unobserve(self) -> None:
        """Stop observing property value changes"""
        # looks like this will be unused, observe property is done via ConsumedThingEvent
        raise NotImplementedError("implement property unobserve per protocol")

    def read_reply(self, message_id: str, timeout: float | int | None = None) -> typing.Any:
        """
        Read the reply of the property get or set which was scheduled with `noblock`.

        Parameters
        ----------
        message_id: str
            id of the request or message (UUID4 as string)
        timeout: float | int | None
            timeout in seconds to wait for the reply, None means wait indefinitely

        Returns
        -------
        typing.Any
            reply of the property get or set
        """
        raise NotImplementedError("implement property read_reply per protocol")


class ConsumedThingEvent:
    # event subscription
    # Dont add class doc otherwise __doc__ in slots will conflict with class variable

    def __init__(
        self,
        resource: EventAffordance,
        logger: logging.Logger,
        owner_inst: typing.Any,
    ) -> None:
        """
        Parameters
        ----------
        resource: EventAffordance
            dataclass object representing the event
        logger: logging.Logger
            logger instance
        owner_inst: typing.Any
            the parent object that owns this event
        """
        self.resource = resource
        self.logger = logger
        self.owner_inst = owner_inst
        self._subscribed = dict()
        # self._sync_callbacks = []
        # self._async_callbacks = []

    def subscribe(
        self,
        callbacks: list[typing.Callable] | typing.Callable,
        asynch: bool = False,
        concurrent: bool = False,
        deserialize: bool = True,
        # create_new_connection: bool = False,
    ) -> None:
        """
        subscribe to the event

        Parameters
        ----------
        callbacks: typing.List[typing.Callable] | typing.Callable
            callback or list of callbacks to add
        asynch: bool
            whether to start an async(-io task) event listener instead of a threaded listener
        concurrent: bool
            - asyncio - if `True`, each callback is scheduled as a separate task, if `False` they are awaited sequentially.
            - threading - if `True`, each callback is called in a separate thread, if `False` they are called sequentially
        deserialize: bool
            if `False`, event payload is passed to the callbacks as raw bytes, if `True` it is deserialized
        """
        op = Operations.observeproperty if isinstance(self.resource, PropertyAffordance) else Operations.subscribeevent
        form = self.resource.retrieve_form(op, None)
        callbacks = callbacks if isinstance(callbacks, (list, tuple)) else [callbacks]
        # if not create_new_connection:
        #   see tag v0.3.2 for logic
        if form is None:
            raise ValueError(f"No form found for {op} operation for {self.resource.name}")
        if asynch:
            get_current_async_loop().call_soon(
                lambda: asyncio.create_task(self.async_listen(form, callbacks, concurrent, deserialize))
            )
        else:
            _thread = threading.Thread(target=self.listen, args=(form, callbacks, concurrent, deserialize), daemon=True)
            _thread.start()

    def unsubscribe(self):
        """unsubscribe from the event"""
        self._subscribed.clear()
        # self._sync_callbacks.clear()
        # self._async_callbacks.clear()

    def listen(self, form: Form, callbacks: list[typing.Callable], concurrent: bool = True, deserialize: bool = True):
        """
        listen to events and call the callbacks
        """
        raise NotImplementedError("implement listen per protocol")

    async def async_listen(
        self, form: Form, callbacks: list[typing.Callable], concurrent: bool = True, deserialize: bool = True
    ):
        """
        listen to events and call the callbacks
        """
        raise NotImplementedError("implement async_listen per protocol")

    def schedule_callbacks(self, callbacks, event_data: typing.Any, concurrent: bool = False) -> None:
        """
        schedule the callbacks to be called with the event data

        Parameters
        ----------
        event_data: typing.Any
            event data to pass to the callbacks
        concurrent: bool
            whether to run each callback in a separate thread
        """
        for cb in callbacks:
            try:
                if not concurrent:
                    cb(event_data)
                else:
                    threading.Thread(target=cb, args=(event_data,)).start()
            except Exception as ex:
                self.logger.error(f"Error occurred in callback {cb}: {ex}")

    async def async_schedule_callbacks(self, callbacks, event_data: typing.Any, concurrent: bool = False) -> None:
        """
        schedule the callbacks to be called with the event data

        Parameters
        ----------
        event_data: typing.Any
            event data to pass to the callbacks
        concurrent: bool
            whether to run each callback in a separate thread
        """
        loop = get_current_async_loop()
        for cb in callbacks:
            try:
                if concurrent:
                    if asyncio.iscoroutinefunction(cb):
                        loop.create_task(cb(event_data))
                    else:
                        loop.run_in_executor(None, cb, event_data)
                elif asyncio.iscoroutinefunction(cb):
                    await cb(event_data)
                else:
                    cb(event_data)
            except Exception as ex:
                self.logger.error(f"Error occurred in callback {cb}: {ex}")

    def add_callbacks(
        self, callbacks: typing.Union[typing.List[typing.Callable], typing.Callable], asynch: bool = False
    ) -> None:
        """
        add callbacks to the event

        Parameters
        ----------
        *callbacks: typing.List[typing.Callable] | typing.Callable
            callback or list of callbacks to add
        """
        raise NotImplementedError(
            "logic error - cannot add callbacks to reuse event subscription. Unsubscribe and resubscribe with new callbacks"
        )
        # for logic, see tag v0.3.2


def raise_local_exception(error_message: typing.Dict[str, typing.Any]) -> None:
    """
    raises an exception on client side using an exception from server, using a mapping based on exception type
    (currently only python built-in exceptions supported). If the exception type is not found, a generic `Exception` is raised.
    Server traceback is added to the exception notes. Client creates its own traceback which is not usually the cause of the error.

    Parameters
    ----------
    error_message: dict[str, Any]
        exception dictionary made by server with following keys - `type`, `message`, `traceback`, `notes`
    """
    if isinstance(error_message, Exception):
        raise error_message from None
    elif isinstance(error_message, dict) and "exception" in error_message.keys():
        error_message = error_message["exception"]
        message = error_message["message"]
        exc = getattr(builtins, error_message["type"], None)
        if exc is None:
            ex = Exception(message)
        else:
            ex = exc(message)
        error_message["traceback"][0] = f"Server {error_message['traceback'][0]}"
        ex.__notes__ = error_message["traceback"][0:-1]
        raise ex from None
    elif isinstance(error_message, str) and error_message in [
        "invokation",
        "execution",
    ]:
        raise TimeoutError(
            f"{error_message[0].upper()}{error_message[1:]} timeout occured. Server did not respond within specified timeout"
        ) from None
    raise RuntimeError("unknown error occurred on server side") from None


@dataclass
class SSE:
    """
    dataclass representing a server sent event and the argument used to invoke event callbacks.

    Attributes
    ----------
    event: str
        event name, defaults to "message"
    data: Any
        event data, defaults to empty string
    id: Optional[str]
        event id, defaults to None
    retry: Optional[int]
        reconnection time in milliseconds, defaults to None, currently unused.
    """

    __slots__ = ("event", "data", "id", "retry")

    def __init__(self):
        self.clear()

    def clear(self):
        """reset to default/empty values"""
        self.event = "message"  # type: str
        self.data = ""  # type: typing.Any
        self.id = None  # type: typing.Optional[str]
        self.retry = None  # type: typing.Optional[int]

    def flush(self) -> typing.Optional[dict]:
        """obtain the event payload as dictionary and reset to default values"""
        if not self.data and self.id is None:
            return None
        payload = {
            "event": self.event,
            "data": self.data,
            "id": self.id,
            "retry": self.retry,
        }
        # reset to default for next event
        self.clear()
        return payload



import threading, asyncio
import logging, multiprocessing, unittest

from hololinked.core.zmq.message import (
    ERROR,
    EXIT,
    OPERATION,
    HANDSHAKE,
    REPLY,
    PreserializedData,
    RequestHeader,
    RequestMessage,
    SerializableData,
)  # client to server
from hololinked.core.zmq.message import (
    TIMEOUT,
    INVALID_MESSAGE,
    ERROR,
    ResponseMessage,
    ResponseHeader,
)  # server to client
from hololinked.core.zmq.brokers import (
    AsyncZMQServer,
    MessageMappedZMQClientPool,
    SyncZMQClient,
    AsyncZMQClient,
)
from hololinked.utils import get_current_async_loop, get_default_logger

try:
    from .utils import TestRunner
    from .test_01_message import MessageValidatorMixin
    from .things.starter import run_zmq_server
    from .things import TestThing
except ImportError:
    from utils import TestRunner
    from test_01_message import MessageValidatorMixin
    from things.starter import run_zmq_server
    from things import TestThing


class TestBrokerMixin(MessageValidatorMixin):
    """Tests Individual ZMQ Server"""

    @classmethod
    def setUpServer(cls):
        cls.server = AsyncZMQServer(id=cls.server_id, logger=cls.logger)

    """
    Base class: BaseZMQ, BaseAsyncZMQ, BaseSyncZMQ
    Servers: BaseZMQServer, AsyncZMQServer, ZMQServerPool
    Clients: BaseZMQClient, SyncZMQClient, AsyncZMQClient, MessageMappedZMQClientPool
    """

    @classmethod
    def setUpClient(cls):
        cls.sync_client = None
        cls.async_client = None

    @classmethod
    def setUpThing(cls):
        cls.thing = TestThing(id=cls.thing_id, logger=cls.logger, remote_accessible_logger=True)

    @classmethod
    def startServer(cls):
        cls._server_thread = threading.Thread(
            target=run_zmq_server, args=(cls.server, cls, cls.done_queue), daemon=True
        )
        cls._server_thread.start()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test ZMQ message brokers {cls.__name__}")
        cls.logger = get_default_logger("test-message-broker", logging.ERROR)
        cls.done_queue = multiprocessing.Queue()
        cls.last_server_message = None
        cls.setUpThing()
        cls.setUpServer()
        cls.setUpClient()
        cls.startServer()


class TestBasicServerAndClient(TestBrokerMixin):
    @classmethod
    def setUpClient(cls):
        super().setUpClient()
        cls.sync_client = SyncZMQClient(
            id=cls.client_id,
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
        )
        cls.client = cls.sync_client

    def test_1_handshake_complete(self):
        """
        Test handshake so that client can connect to server. Once client connects to server,
        verify a ZMQ internal monitoring socket is available.
        """
        self.client.handshake()
        self.assertTrue(self.client._monitor_socket is not None)
        self.assertTrue(self.client._monitor_socket in self.client.poller)
        # both directions
        # HANDSHAKE = 'HANDSHAKE' # 1 - find out if the server is alive

    def test_2_message_contract_types(self):
        """
        Once composition is checked, check different message types
        """
        # message types
        request_message = RequestMessage.craft_from_arguments(
            receiver_id=self.server_id,
            sender_id=self.client_id,
            thing_id=self.thing_id,
            objekt="some_prop",
            operation="readProperty",
        )

        async def handle_message_types_server():
            # server to client
            # REPLY = b'REPLY' # 4 - response for operation
            # TIMEOUT = b'TIMEOUT' # 5 - timeout message, operation could not be completed
            # EXCEPTION = b'EXCEPTION' # 6 - exception occurred while executing operation
            # INVALID_MESSAGE = b'INVALID_MESSAGE' # 7 - invalid message
            await self.server._handle_timeout(request_message, timeout_type="execution")  # 5
            await self.server._handle_invalid_message(request_message, SerializableData(Exception("test")))  # 7
            await self.server._handshake(request_message)  # 1
            await self.server._handle_error_message(request_message, Exception("test"))  # 6
            await self.server.async_send_response(request_message)  # 4
            await self.server.async_send_response_with_message_type(
                request_message, ERROR, SerializableData(Exception("test"))
            )  # 6

        get_current_async_loop().run_until_complete(handle_message_types_server())

        """
        message types

        both directions
        HANDSHAKE = b'HANDSHAKE' # 1 - taken care by test_1...
        
        client to server 
        OPERATION = b'OPERATION' 2 - taken care by test_2_... # operation request from client to server
        EXIT = b'EXIT' # 3 - taken care by test_7... # exit the server
        
        server to client
        REPLY = b'REPLY' # 4 - response for operation
        TIMEOUT = b'TIMEOUT' # 5 - timeout message, operation could not be completed
        EXCEPTION = b'EXCEPTION' # 6 - exception occurred while executing operation
        INVALID_MESSAGE = b'INVALID_MESSAGE' # 7 - invalid message
        SERVER_DISCONNECTED = 'EVENT_DISCONNECTED' not yet tested # socket died - zmq's builtin event
        
        peer to peer
        INTERRUPT = b'INTERRUPT' not yet tested # interrupt a socket while polling 
        """

        msg = self.client.recv_response(request_message.id)
        self.assertEqual(msg.type, TIMEOUT)
        self.validate_response_message(msg)

        msg = self.client.recv_response(request_message.id)
        self.assertEqual(msg.type, INVALID_MESSAGE)
        self.validate_response_message(msg)

        msg = self.client.socket.recv_multipart()  # handshake dont come as response
        response_message = ResponseMessage(msg)
        self.assertEqual(response_message.type, HANDSHAKE)
        self.validate_response_message(response_message)

        msg = self.client.recv_response(request_message.id)
        self.assertEqual(msg.type, ERROR)
        self.validate_response_message(msg)

        msg = self.client.recv_response(request_message.id)
        self.assertEqual(msg.type, REPLY)
        self.validate_response_message(msg)

        msg = self.client.recv_response(request_message.id)
        # custom crafted explicitly to be ERROR
        self.assertEqual(msg.type, ERROR)
        self.validate_response_message(msg)

        self.client.handshake()

    def test_3_verify_polling(self):
        """
        Test if polling may be stopped and started again
        """

        async def verify_poll_stopped(self: TestBasicServerAndClient) -> None:
            await self.server.poll_requests()
            self.server.poll_timeout = 1000
            await self.server.poll_requests()
            self.done_queue.put(True)

        async def stop_poll(self: TestBasicServerAndClient) -> None:
            await asyncio.sleep(0.1)
            self.server.stop_polling()
            await asyncio.sleep(0.1)
            self.server.stop_polling()

        # When the above two functions running,
        # we dont send a message as the thread is also running
        get_current_async_loop().run_until_complete(asyncio.gather(*[verify_poll_stopped(self), stop_poll(self)]))

        self.assertTrue(self.done_queue.get())
        self.assertEqual(self.server.poll_timeout, 1000)
        self.client.handshake()

    @classmethod
    def tearDownClass(cls):
        """
        Test if exit reaches to server
        """
        # EXIT = b'EXIT' # 7 - exit the server
        request_message = RequestMessage.craft_with_message_type(
            receiver_id=cls.server_id, sender_id=cls.client_id, message_type=EXIT
        )
        cls.client.socket.send_multipart(request_message.byte_array)

        # TODO - fix the following, somehow socket is not closing fully,
        # although we have previously tested this and its known to work.
        # try:
        #     cls.client.recv_response(message_id=b'not-necessary')
        #     assert False, "Expected ConnectionAbortedError"
        # except ConnectionAbortedError as ex:
        #     assert str(ex).startswith(f"server disconnected for {cls.client_id}"), f"Unexpected error message: {str(ex)}"

        done = cls.done_queue.get(timeout=3)
        if done:
            cls._server_thread.join()
        else:
            print("Server did not properly process exit request")
        super().tearDownClass()

    # TODO
    # peer to peer
    # INTERRUPT = b'INTERRUPT' # interrupt a socket while polling
    # first test the length


class TestAsyncZMQClient(TestBrokerMixin):
    @classmethod
    def setUpClient(cls):
        cls.async_client = AsyncZMQClient(
            id=cls.client_id,
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
        )
        cls.client = cls.async_client

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_1_handshake_complete(self):
        """
        Test handshake so that client can connect to server. Once client connects to server,
        verify a ZMQ internal monitoring socket is available.
        """

        async def test():
            self.client.handshake()
            await self.client.handshake_complete()
            self.assertTrue(self.client._monitor_socket is not None)
            self.assertTrue(self.client._monitor_socket in self.client.poller)

        get_current_async_loop().run_until_complete(test())
        # both directions
        # HANDSHAKE = 'HANDSHAKE' # 1 - find out if the server is alive

    def test_2_message_contract_types(self):
        """
        Once composition is checked, check different message types
        """
        # message types
        request_message = RequestMessage.craft_from_arguments(
            receiver_id=self.server_id,
            sender_id=self.client_id,
            thing_id=self.thing_id,
            objekt="some_prop",
            operation="readProperty",
        )

        async def handle_message_types_server():
            # server to client
            # REPLY = b'REPLY' # 4 - response for operation
            # TIMEOUT = b'TIMEOUT' # 5 - timeout message, operation could not be completed
            # EXCEPTION = b'EXCEPTION' # 6 - exception occurred while executing operation
            # INVALID_MESSAGE = b'INVALID_MESSAGE' # 7 - invalid message
            await self.server._handle_timeout(request_message, timeout_type="invokation")  # 5
            await self.server._handle_invalid_message(request_message, SerializableData(Exception("test1")))
            await self.server._handshake(request_message)
            await self.server._handle_error_message(request_message, Exception("test2"))
            await self.server.async_send_response(request_message)
            await self.server.async_send_response_with_message_type(
                request_message, ERROR, SerializableData(Exception("test3"))
            )

        async def handle_message_types_client():
            """
            message types
            both directions
            HANDSHAKE = b'HANDSHAKE' # 1 - taken care by test_1...

            client to server
            OPERATION = b'OPERATION' 2 - taken care by test_2_... # operation request from client to server
            EXIT = b'EXIT' # 3 - taken care by test_7... # exit the server

            server to client
            REPLY = b'REPLY' # 4 - response for operation
            TIMEOUT = b'TIMEOUT' # 5 - timeout message, operation could not be completed
            EXCEPTION = b'EXCEPTION' # 6 - exception occurred while executing operation
            INVALID_MESSAGE = b'INVALID_MESSAGE' # 7 - invalid message
            SERVER_DISCONNECTED = 'EVENT_DISCONNECTED' not yet tested # socket died - zmq's builtin event

            peer to peer
            INTERRUPT = b'INTERRUPT' not yet tested # interrupt a socket while polling
            """
            msg = await self.client.async_recv_response(request_message.id)
            self.assertEqual(msg.type, TIMEOUT)
            self.validate_response_message(msg)

            msg = await self.client.async_recv_response(request_message.id)
            self.assertEqual(msg.type, INVALID_MESSAGE)
            self.validate_response_message(msg)

            msg = await self.client.socket.recv_multipart()  # handshake don't come as response
            response_message = ResponseMessage(msg)
            self.assertEqual(response_message.type, HANDSHAKE)
            self.validate_response_message(response_message)

            msg = await self.client.async_recv_response(request_message.id)
            self.assertEqual(msg.type, ERROR)
            self.validate_response_message(msg)

            msg = await self.client.async_recv_response(request_message.id)
            self.assertEqual(msg.type, REPLY)
            self.validate_response_message(msg)

            msg = await self.client.async_recv_response(request_message.id)
            self.assertEqual(msg.type, ERROR)
            self.validate_response_message(msg)

        # exit checked separately at the end
        get_current_async_loop().run_until_complete(
            asyncio.gather(*[handle_message_types_server(), handle_message_types_client()])
        )

    @classmethod
    def tearDownClass(cls):
        """
        Test if exit reaches to server
        """
        # EXIT = b'EXIT' # 7 - exit the server
        request_message = RequestMessage.craft_with_message_type(
            receiver_id=cls.server_id, sender_id=cls.client_id, message_type=EXIT
        )
        cls.client.socket.send_multipart(request_message.byte_array)
        done = cls.done_queue.get(timeout=3)

        # TODO - check server disconnected like previous test

        if done:
            cls._server_thread.join()
        else:
            print("Server did not properly process exit request")
        super().tearDownClass()


class TestMessageMappedClientPool(TestBrokerMixin):
    @classmethod
    def setUpClient(cls):
        cls.client = MessageMappedZMQClientPool(
            id="client-pool",
            client_ids=[cls.client_id],
            server_ids=[cls.server_id],
            logger=cls.logger,
            handshake=False,
        )
        cls.client._client_to_thing_map[cls.client_id] = cls.thing_id
        cls.client._thing_to_client_map[cls.thing_id] = cls.client_id

    def test_1_handshake_complete(self):
        """
        Test handshake so that client can connect to server. Once client connects to server,
        verify a ZMQ internal monitoring socket is available.
        """

        async def test():
            self.client.handshake()
            await self.client.handshake_complete()
            for client in self.client.pool.values():
                self.assertTrue(client._monitor_socket is not None)
                self.assertTrue(client._monitor_socket in self.client.poller)

        get_current_async_loop().run_until_complete(test())
        # both directions
        # HANDSHAKE = 'HANDSHAKE' # 1 - find out if the server is alive

    def test_2_message_contract_types(self):
        """
        Once composition is checked, check different message types
        """
        # message types
        request_message = RequestMessage.craft_from_arguments(
            receiver_id=self.server_id,
            sender_id=self.client_id,
            thing_id=self.thing_id,
            objekt="some_prop",
            operation="readProperty",
        )

        async def handle_message_types():
            """
            message types
            both directions
            HANDSHAKE = b'HANDSHAKE' # 1 - taken care by test_1...

            client to server
            OPERATION = b'OPERATION' 2 - taken care by test_2_... # operation request from client to server
            EXIT = b'EXIT' # 3 - taken care by test_7... # exit the server

            server to client
            REPLY = b'REPLY' # 4 - response for operation
            TIMEOUT = b'TIMEOUT' # 5 - timeout message, operation could not be completed
            EXCEPTION = b'EXCEPTION' # 6 - exception occurred while executing operation
            INVALID_MESSAGE = b'INVALID_MESSAGE' # 7 - invalid message
            SERVER_DISCONNECTED = 'EVENT_DISCONNECTED' not yet tested # socket died - zmq's builtin event

            peer to peer
            INTERRUPT = b'INTERRUPT' not yet tested # interrupt a socket while polling
            """
            self.client.start_polling()

            self.client.events_map[request_message.id] = self.client.event_pool.pop()
            await self.server._handle_timeout(request_message, timeout_type="invokation")  # 5
            msg = await self.client.async_recv_response(self.thing_id, request_message.id)
            self.assertEqual(msg.type, TIMEOUT)
            self.validate_response_message(msg)

            self.client.events_map[request_message.id] = self.client.event_pool.pop()
            await self.server._handle_invalid_message(request_message, SerializableData(Exception("test")))
            msg = await self.client.async_recv_response(self.thing_id, request_message.id)
            self.assertEqual(msg.type, INVALID_MESSAGE)
            self.validate_response_message(msg)

            self.client.events_map[request_message.id] = self.client.event_pool.pop()
            await self.server._handshake(request_message)
            msg = await self.client.pool[self.client_id].socket.recv_multipart()  # handshake don't come as response
            response_message = ResponseMessage(msg)
            self.assertEqual(response_message.type, HANDSHAKE)
            self.validate_response_message(response_message)

            self.client.events_map[request_message.id] = self.client.event_pool.pop()
            await self.server.async_send_response(request_message)
            msg = await self.client.async_recv_response(self.thing_id, request_message.id)
            self.assertEqual(msg.type, REPLY)
            self.validate_response_message(msg)

            self.client.events_map[request_message.id] = self.client.event_pool.pop()
            await self.server.async_send_response_with_message_type(
                request_message, ERROR, SerializableData(Exception("test"))
            )
            msg = await self.client.async_recv_response(self.thing_id, request_message.id)
            self.assertEqual(msg.type, ERROR)
            self.validate_response_message(msg)

            self.client.stop_polling()

        # exit checked separately at the end
        get_current_async_loop().run_until_complete(asyncio.gather(*[handle_message_types()]))

    def test_3_verify_polling(self):
        """
        Test if polling may be stopped and started again
        """

        async def verify_poll_stopped(self: "TestMessageMappedClientPool") -> None:
            await self.client.poll_responses()
            self.client.poll_timeout = 1000
            await self.client.poll_responses()
            self.done_queue.put(True)

        async def stop_poll(self: "TestMessageMappedClientPool") -> None:
            await asyncio.sleep(0.1)
            self.client.stop_polling()
            await asyncio.sleep(0.1)
            self.client.stop_polling()

        # When the above two functions running,
        # we dont send a message as the thread is also running
        get_current_async_loop().run_until_complete(asyncio.gather(*[verify_poll_stopped(self), stop_poll(self)]))
        self.assertTrue(self.done_queue.get())
        self.assertEqual(self.client.poll_timeout, 1000)

    @classmethod
    def tearDownClass(cls):
        """
        Test if exit reaches to server
        """
        # EXIT = b'EXIT' # 7 - exit the server
        request_message = RequestMessage.craft_with_message_type(
            receiver_id=cls.server_id, sender_id=cls.client_id, message_type=EXIT
        )
        cls.client[cls.client_id].socket.send_multipart(request_message.byte_array)
        done = cls.done_queue.get(timeout=3)
        if done:
            cls._server_thread.join()
        else:
            print("Server did not process exit message correctly")
        super().tearDownClass()


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestBasicServerAndClient))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestAsyncZMQClient))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestMessageMappedClientPool))
    return suite


if __name__ == "__main__":
    runner = TestRunner()
    runner.run(load_tests(unittest.TestLoader(), None, None))



from typing import Any, Optional


def get_summary(docs: str) -> Optional[str]:
    """Return the first line of the dosctring of an object

    :param obj: Any Python object
    :returns: str: First line of object docstring

    """
    if docs:
        return docs.partition("\n")[0].strip()
    else:
        return ""


def get_zmq_unique_identifier_from_event_affordance(affordance: Any) -> Optional[str]:
    """Return the unique identifier for a ZMQ object

    :param obj: Any Python object
    :returns: str: Unique identifier for the object

    """
    return f"{affordance.thing_id}/{affordance.name}"
